﻿using EA;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.DirectoryServices.ActiveDirectory;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Runtime.Remoting.Contexts;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using static System.Windows.Forms.VisualStyles.VisualStyleElement.Window;

namespace CimSyntaxGen.AsciiDoc
{
    public class AsciiDocGenerator
    {
        private EA.Repository repo;
        private static string ASSEMBLY = "assembly";
        private static string CONTEXTUAL = "contextual";
        public static string IMAGE_DIR = "images";
        public static string LBR = " +\n"; // hard line break
        private EA.Element rootElem;
        List<string> notPureClassStereotypes = new List<string>();
        Dictionary<string, List<long>> TypeList = new Dictionary<string, List<long>>();
        ConstantDefinition CD = new ConstantDefinition();

        //private Dictionary<int, EA.Element> classesInPackage = new Dictionary<int, EA.Element>();

        public AsciiDocGenerator(EA.Repository repo)
        {
            this.repo = repo;
        }

        public string CreateDocu(EA.Package pack, string outputDir)
        {
            string directives = GetAsciiDocDirectives();
            string packNameLowerCase = pack.Name.ToLower();
            if (packNameLowerCase.Contains(CONTEXTUAL))
            {
                return directives + CreateContextualDocu(pack, outputDir);
            } else if (packNameLowerCase.Contains(ASSEMBLY))
            {
                return directives + CreateAssemblyDocu(pack);
            } else // might have two or more packages, at least one for contextual and the other for assembly model
            {
                EA.Package contextualPkg = null;
                EA.Package assemblyPgk = null;
                foreach (EA.Package pkg in pack.Packages)
                {
                    if (pkg.Name.ToLower().Contains(CONTEXTUAL)) contextualPkg = pkg;
                    if (pkg.Name.ToLower().Contains(ASSEMBLY)) assemblyPgk = pkg;
                }
                if (contextualPkg != null && assemblyPgk != null)
                {
                    string contextualDocu = CreateContextualDocu(contextualPkg, outputDir);
                    string assemblyDocu = CreateAssemblyDocu(assemblyPgk);
                    return directives + contextualDocu + assemblyDocu;
                }
            }
            return "";
        }

        public void SaveCoverpageImage(string outputDir)
        {
            string FILENAME = "entose_logo-h-pos.png";
            Bitmap logoImg = FileManager.GetImageByName("CimSyntaxGen.Option.", FILENAME);
            if (logoImg != null)
            {
                string imageDirName = outputDir + IMAGE_DIR;
                if (!Directory.Exists(imageDirName))
                {
                    Directory.CreateDirectory(imageDirName);
                }
                try
                {
                    logoImg.Save(imageDirName + "\\" + FILENAME, System.Drawing.Imaging.ImageFormat.Png);
                }
                catch (Exception e)
                {
                    ErrorCodes.ShowException(ErrorCodes.ERROR_053, e);
                }
            }
        }

        public string GetCoverIntroAsciiDoc(string title)
        {
            string TAG_TITLE = "$TITLE$";
            string TAG_DATE = "$CURRENTDATE$";
            string coverIntro = FileManager.GetTextByName("CimSyntaxGen.Resources.", "CoverAndIntro.adoc");
            if(coverIntro != null) 
            {
                return coverIntro.Replace(TAG_TITLE, title).Replace(TAG_DATE, DateTime.Now.ToString("yyyy-MM-dd"));
            }
            return "";
        }

        public string GetCoverIntroAsciiDoc_61970(string title)
        {
            string TAG_TITLE = "$TITLE$";
            string TAG_DATE = "$CURRENTDATE$";
            string coverIntro = FileManager.GetTextByName("CimSyntaxGen.Resources.", "CoverAndIntro_61970.adoc");
            if (coverIntro != null)
            {
                return coverIntro.Replace(TAG_TITLE, title).Replace(TAG_DATE, DateTime.Now.ToString("yyyy-MM-dd"));
            }
            return "";
        }


        public void SaveAsciiDoc(string fileName, string dirName, string asciiDoc)
        {
            if (asciiDoc == null || asciiDoc.Length == 0) return;
            string fullPath = dirName + fileName;
            try
            {
                System.IO.File.WriteAllText(fullPath, asciiDoc);
            } catch (Exception e)
            {
                ErrorCodes.ShowException(ErrorCodes.ERROR_050, e);
            }
        }

        public string GetAsciiDocDirectives()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(":imagesdir: ./" + IMAGE_DIR + "\n");
            sb.Append(":figure-number: 0\n");
            sb.Append(":table-number: 0\n");
            sb.Append(":listing-number: 0\n");
            sb.Append(":figure-caption: Figure {counter:chapter}.\n");
            sb.Append(":table-caption: Table {chapter}.\n");
            sb.Append(":listing-caption: Listing {chapter}.\n");
            sb.Append("\n\n");
            return sb.ToString();
        }

        private Dictionary<int, EA.Element> GetClassesInPackage(EA.Package pack)
        {
            Dictionary<int, EA.Element> classes = new Dictionary<int, EA.Element>();
            foreach (EA.Element elem in pack.Elements)
            {
                classes.Add(elem.ElementID, elem);
            }
            return classes;
        }

        public string SaveDiagram(EA.Package pack, string outputDir)
        {
            EA.Collection diagramCollection = repo.GetPackageByID(pack.PackageID).Diagrams;
            string diagFileName = null;
            if (diagramCollection.Count > 0)
            {
                EA.Diagram diag = (EA.Diagram)diagramCollection.GetAt(0);
                diagFileName = diag.Name.Replace(" ", "_") + ".png";
                string diagramGUID = diag.DiagramGUID;
                EA.Project project = repo.GetProjectInterface();
                string imageDirName = outputDir + IMAGE_DIR;
                if (!Directory.Exists(imageDirName))
                {
                    Directory.CreateDirectory(imageDirName);
                }
                project.PutDiagramImageToFile(diagramGUID, imageDirName + "\\" + diagFileName, 1);
            }
            return diagFileName;
        }
        private string CreateContextualDocu(EA.Package pack, string outputDir)
        {
            TypeList.Add("datatype", new List<long>());
            TypeList.Add("enumeration", new List<long>());
            TypeList.Add("compound", new List<long>());
            TypeList.Add("primitive", new List<long>());

            this.rootElem = GetRootElement(pack);
            if (this.rootElem == null)
            {
                ErrorCodes.ShowError(ErrorCodes.ERROR_051);
                return "";
            }

            StringBuilder sb = new StringBuilder();
            AsciiDocStyle asciiDocStyle = new AsciiDocStyle();
            sb.Append(asciiDocStyle.GetHeader(pack.Name, 0));
            sb.Append(asciiDocStyle.GetHeader("Overview of the model", 1));
            sb.Append("The following diagram shows the model:\n");

            string diagramFileName = SaveDiagram(pack, outputDir);
            AsciiDocImage asciiDocImage = new AsciiDocImage(pack.Name, diagramFileName, 1, -1, -1);
            sb.Append(asciiDocImage.GetImage());

            sb.Append(asciiDocStyle.GetHeader("IsBasedOn relationships:", 1));
            sb.Append("The following table shows the traceability dependency of the classes used in this package towards the upper level.\n\n");

            AsciiDocTable asciiDocTable = new AsciiDocTable("Table - IsBasedOn Dependency");
            asciiDocTable.SetHeaderRow(new List<string>() { "Name", "Complete IsBasedOn Path" });
            List<string[]> rowContent = GetIsBasedOnPaths(pack);
            foreach (string[] row in rowContent)
            {
                asciiDocTable.AddRow(new List<string>() { row[0], row[1] });
            }
            sb.Append("\n");
            sb.Append(asciiDocTable.GetTable("1, 4", 100));
            sb.Append(asciiDocStyle.GetHeader("Detailed " + pack.Name, 1));
            sb.Append(GetClassDetails(pack, this.rootElem));
            sb.Append("\n\n");

            List<EA.Element> profileClasses = GetTypedElementsWithoutRoot(pack, "Class");
            //ProfileClasses.Sort(RDFutilitaires.CompareElementsbyName);
            List<EA.Element> classlist = profileClasses.FindAll(IsClassWithoutStereotype);
            classlist.Sort(RDFutilitaires.CompareElementsbyName);

            foreach (EA.Element elem in classlist)
            {
                if (elem.ElementID == this.rootElem.ElementID) continue;
                sb.Append(GetClassDetails(pack, elem));
                sb.Append("\n\n");
            }

            SelectStereotypedClasses(profileClasses);
            // primitives, compounds, datatypes

            return sb.ToString();
        }

        // Assumption: There's just one root element in the package.
        public EA.Element GetRootElement(EA.Package pkg)
        {
            foreach (EA.Element elem in pkg.Elements)
            {
                if (elem.IsActive) return elem;
            }
            return null;
        }

        public static List<EA.Element> GetTypedElementsWithoutRoot(EA.Package package, string type)
        {
            List<EA.Element> resultList = new List<EA.Element>();
            foreach (EA.Element elem in package.Elements)
            {
                if (elem.IsActive) continue;
                if (elem.Type.Equals(type) && !resultList.Contains(elem)) resultList.Add(elem);
            }
            foreach (EA.Package pkg in package.Packages)
            {
                resultList.AddRange(GetTypedElementsWithoutRoot(pkg, type));
            }
            return resultList;
        }


        private string GetClassDetails(EA.Package pack, EA.Element elem)
        {
            Dictionary<int, EA.Element> classesInPkg = GetClassesInPackage(pack);
            StringBuilder sb = new StringBuilder();
            AsciiDocStyle asciiDocStyle = new AsciiDocStyle();
            sb.Append(asciiDocStyle.GetHeader(elem.Name, 2));
            sb.Append(elem.Notes + LBR);

            // Attributes
            sb.Append("The following table shows all attributes of " + elem.Name + ":" + LBR + "\n");

            AsciiDocTable asciiDocTable = new AsciiDocTable("Table of Attributes of " + pack.Name + ":" + elem.Name);
            asciiDocTable.SetHeaderRow(new List<string>() { "Order", "Mult.", "Attribute Name / Type", "Description" });
            int order = 1;
            foreach (EA.Attribute attr in elem.Attributes)
            {
                asciiDocTable.AddRow(new List<string>() { order + "",
                                                         "[" + attr.LowerBound + ".." + attr.UpperBound + "]",
                                                         attr.Name + " /" + LBR + attr.Type,
                                                         attr.Notes});
                order++;
            }
            sb.Append(asciiDocTable.GetTable("1, 2, 3, 5", 100));

            // Association ends
            List<EA.Connector> connList = GetRelevantConnectors(elem);
            sb.Append("The following table shows all association ends of " + elem.Name + " with other classes." + LBR + "\n");
            AsciiDocTable asciiDocAssocTable = new AsciiDocTable("Table of association ends of  " + pack.Name + ":" + elem.Name);
            asciiDocAssocTable.SetHeaderRow(new List<string>() { "Order", "Mult.", "Class Name / Role", "Description" });
            EA.Element otherElem = null;
            EA.ConnectorEnd otherConnEnd = null;
            if (connList.Count > 0)
            {
                foreach (EA.Connector conn in connList)
                {
                    if (conn.SupplierID == elem.ElementID)
                    {
                        otherElem = repo.GetElementByID(conn.ClientID);
                        otherConnEnd = conn.ClientEnd;
                    }
                    else if (conn.ClientID == elem.ElementID)
                    {
                        otherElem = repo.GetElementByID(conn.SupplierID);
                        otherConnEnd = conn.SupplierEnd;
                    }
                    if (classesInPkg.ContainsKey(otherElem.ElementID))
                    {
                        string note = otherConnEnd.RoleNote + LBR + "Association Based On: " + LBR + GetAssocIsBasedOn(conn)[0] + LBR +
                                       "-----" + LBR + GetAssocIsBasedOn(conn)[1];
                        asciiDocAssocTable.AddRow(new List<string>() { order + "",
                                       "[" + otherConnEnd.Cardinality + "]",
                                       otherElem.Name + " /" + LBR + otherConnEnd.Role,
                                       note});
                        order++;
                    }
                }
                sb.Append(asciiDocAssocTable.GetTable("1, 2, 3, 5", 100));
            }
            return sb.ToString();
        }

        public string GetPackageContainmentPath(EA.Element el)
        {
            string ret = "";
            try
            {
                if (el != null)
                {
                    List<String> elemPath = new List<String>();
                    DirectPackageContainmentPath(this.repo.GetPackageByID(el.PackageID), elemPath);
                    elemPath.Reverse();
                    if (elemPath[0] == "Model") elemPath.Remove("Model");
                    StringBuilder qualifiedElemPath = new StringBuilder();
                    elemPath.ForEach(delegate (String name)
                    {
                        qualifiedElemPath.Append("::" + name);
                    });

                    ret = qualifiedElemPath.ToString().TrimStart(new char[] { ':' });
                    return ret;
                }
                else
                {
                    MessageBox.Show("Error GetPackageBasedOnContainmentPath el is null!");
                    return ret;
                }
            }
            catch (Exception e)
            {
                MessageBox.Show("Error on GetPackageBasedOnContainmentPath!");
            }
            return ret;
        }

        public string GetAssocAsciiDoc(EA.Element elt, EA.Package pack, EA.Connector con)
        {
            CIMSchemaGenerator.ConexesID = 0;
            CIMSchemaGenerator.Connexion connex = new CIMSchemaGenerator.Connexion(con);
            EA.ConnectorEnd otherend;
            EA.ConnectorEnd myend;
            EA.Element otherendelt;

            if (connex.Source_element.ElementID == elt.ElementID)
            {// the element is the Source
                otherend = connex.Target;
                otherendelt = connex.Target_element;
                myend = connex.Source;
            }
            else
            {
                otherend = connex.Source;
                otherendelt = connex.Source_element;
                myend = connex.Target;
            }
            string cardinality = otherend.Cardinality;
            bool eligible = false;
            if (cardinality == "1") cardinality = "1..1";
            if (myend.Aggregation != 0)

            {// the other end is eligible as member
                eligible = true;
            }
            StringBuilder sb = new StringBuilder();
            if (eligible && (otherend.Role != ""))
            {
                {// the association creates an effective memmber
                    sb.Append("\nThe following table shows all association ends of " + elt.Name + " with other classes.");
                    AsciiDocTable asciiDocTable = new AsciiDocTable("Table - Association ends of " + pack.Name + "::" + elt.Name + " with other classes");
                    asciiDocTable.SetHeaderRow(new List<string>() { "Mult.", "Class name / Role", "Description" });
                    string[] assocBasedOn = GetAssocIsBasedOn(con);
                    asciiDocTable.AddRow(new List<string>() { "[" + cardinality + "]",
                                                         elt.Name + "." + otherend.Role + "\">" + otherendelt.Name + " / " + AsciiDocGenerator.LBR + otherend.Role,
                                                         otherend.RoleNote.Replace("\n", AsciiDocGenerator.LBR) +
                                                         AsciiDocGenerator.LBR + assocBasedOn[0] +
                                                         AsciiDocGenerator.LBR + "--------" +
                                                         AsciiDocGenerator.LBR + assocBasedOn[1]});
                    sb.Append(asciiDocTable.GetTable("1, 2, 4", 100));
                }
            }
            return sb.ToString();
        }


        public bool IsConEligible(EA.Element el, EA.Connector con)
        {
            if ((con.Type != "Association") && (con.Type != "Aggregation")) return false;
            if ((el.ElementID == con.SupplierID) && (con.SupplierEnd.Aggregation == 1)
                && (con.ClientEnd.Role != ""))
            {
                return true;
            }
            if ((el.ElementID == con.ClientID) && (con.ClientEnd.Aggregation == 1)
               && (con.SupplierEnd.Role != ""))
            {
                return true;
            }
            return false;
        }



        // Calculates the Package containment from a given Package
        // The result is in the form of a string list in reverse order
        private void DirectPackageContainmentPath(EA.Package pkg, List<String> path)
        {
            path.Add(pkg.Name);
            if (pkg.ParentID != 0) // up to top
            {
                DirectPackageContainmentPath(this.repo.GetPackageByID(pkg.ParentID), path);
            }
        }

        public string GenerateNoteAsciiDoc(EA.Element elt, EA.Connector con, string isAbstract)
        {
            long eid;
            if (elt.ElementID == con.ClientID)
            {
                eid = con.SupplierID; // note is on the otherend of the notelink
            }
            else
            {
                eid = con.ClientID; // note is at the source end of the notelink
            }
            string note = "";
            if (eid != 0)
            {
                note = Main.Repo.GetElementByID((int)eid).Notes;
            }
            //string content = "-" + note.Replace("\n", AsciiDocGenerator.LBR);
            string content = "-" + note.Replace("\n", "");
            return content;
        }



        public string[] GetAssocIsBasedOn(EA.Connector conn)
        {
            string IBOGuid = RDFutilitaires.getBasedOnConnectoryGUID(repo, conn.ConnectorGUID);
            EA.Connector IBOcon = repo.GetConnectorByGuid(IBOGuid);
            string[] text = { "", "" };
            if (IBOcon != null)
            {
                EA.Element source_element = repo.GetElementByID((int)IBOcon.ClientID);
                EA.Element target_element = repo.GetElementByID((int)IBOcon.SupplierID);
                EA.ConnectorEnd source = IBOcon.ClientEnd;
                EA.ConnectorEnd target = IBOcon.SupplierEnd;

                text[0] = repo.GetPackageByID((int)source_element.PackageID).Name + "::" + source_element.Name + "." + source.Role;
                text[0] = text[0] + "[" + source.Cardinality + "]";
                text[1] = repo.GetPackageByID((int)target_element.PackageID).Name + "::" + target_element.Name + "." + target.Role;
                text[1] = text[1] + "[" + target.Cardinality + "]";
            }
            else
            {
                MessageBox.Show("WARNING there is a problem in the model with the association roles: ClientRole = " +
                    conn.ClientEnd.Role + ", " + "SupplierRole = " + conn.SupplierEnd.Role + "\n Model is not built with CimSyntaxGen!");
            }
            return text;
        }


        private List<EA.Connector> GetRelevantConnectors(EA.Element elem)
        {
            List<EA.Connector> connList = new List<EA.Connector>();
            foreach (EA.Connector conn in elem.Connectors)
            {
                if ((elem.ElementID == conn.ClientID) && (conn.ClientEnd.Aggregation > 0))
                {
                    connList.Add(conn);
                } else if ((elem.ElementID == conn.SupplierID) && (conn.SupplierEnd.Aggregation > 0))
                {
                    connList.Add(conn);
                }
            }
            return connList;
        }


        /*
        private List<string[]> GetIsBasedOnPaths(EA.Package pack)
        {
            List<string[]> pathsList = new List<string[]>();
            foreach (EA.Element elem in pack.Elements)
            {
                int otherEndID = -1;
                foreach (EA.Connector conn in elem.Connectors)
                {
                    if(conn.Stereotype.Equals("IsBasedOn"))
                    {
                        if (conn.SupplierID == elem.ElementID) otherEndID = conn.ClientID;
                        else if (conn.ClientID == elem.ElementID) otherEndID = conn.SupplierID;
                        break;
                    }
                }

                if(otherEndID >= 0)
                {
                    EA.Element basedOnElem = repo.GetElementByID(otherEndID);
                    if (basedOnElem != null)
                    {
                        pathsList.Add(new string[] { basedOnElem.Name, GetFullPackagePath(basedOnElem) });
                    }
                }
            }
            return pathsList;
        }
        */

        private List<string[]> GetIsBasedOnPaths(EA.Package pack)
        {
            List<string[]> pathsList = new List<string[]>();
            foreach (EA.Element elem in pack.Elements)
            {
                EA.Element finalBasedOnElem = GetFinalIsBasedOnElement(elem);
                pathsList.Add(new string[] { finalBasedOnElem.Name, GetFullPackagePath(finalBasedOnElem) });
            }
            return pathsList;
        }

        private EA.Element GetFinalIsBasedOnElement(EA.Element elem)
        {
            EA.Element retElem = elem;
            foreach (EA.TaggedValue tagValue in elem.TaggedValues)
            {
                if (tagValue.Name.Equals("GUIDBasedOn"))
                {
                    string guidBasedOn = tagValue.Value;
                    if (guidBasedOn != null && guidBasedOn.Length > 0)
                    {
                        EA.Element basedOnElem = repo.GetElementByGuid(guidBasedOn);
                        if (basedOnElem == retElem) { return retElem; }
                        else retElem = GetFinalIsBasedOnElement(basedOnElem);
                    }
                }
            }
            return retElem;
        }

        private string GetFullPackagePath(EA.Element elem)
        {
            string packPath = elem.Name;
            EA.Package pkg = repo.GetPackageByID(elem.PackageID);
            return GetPackagePathElem(pkg, packPath);
        }

        private string GetPackagePathElem(EA.Package pkg, string packPath)
        {
            if (pkg != null)
            {
                if (pkg.ParentID > 0) // if 0 then pkg is a model
                {
                    packPath = pkg.Name + "::" + packPath;
                    EA.Package pkg2 = repo.GetPackageByID(pkg.ParentID);
                    return GetPackagePathElem(pkg2, packPath);
                }
                else
                {
                    return packPath;
                }
            }
            return "";
        }

        private string CreateAssemblyDocu(EA.Package pack)
        {
            StringBuilder sb = new StringBuilder();

            return sb.ToString();
        }

        private string GetPrimitiveAsciiDoc(EA.Element elem)
        {
            StringBuilder sb = new StringBuilder();
            AsciiDocStyle asciiDocStyle = new AsciiDocStyle();

            if (elem.Name != null && elem.Name.Length > 0)
            {
                sb.Append(asciiDocStyle.GetHeader(elem.Name + " primitive", 3));
                sb.Append(elem.Notes + LBR);
            }
            return sb.ToString();
        }


        private string GetCompoundAsciiDoc(EA.Element elem)
        {
            return GetAsciiDocWithAttributesTable(elem, "compound");
        }

        private string GetDatatypeAsciiDoc(EA.Element elem)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(GetAsciiDocWithAttributesTable(elem, "datatype"));
            sb.Append(LBR);
            foreach (EA.Attribute attr in elem.Attributes)
            {
                sb.Append(GetAsciiDocWithContraintsTable(attr));
            }
            return sb.ToString();
        }

        private string GetAsciiDocWithContraintsTable(EA.Attribute at)
        {
            StringBuilder sb = new StringBuilder();
            bool constraintsExist = false;
            sb.Append("Restrictions applied to the attributes of " + at.Name + ":" + LBR);
            AsciiDocTable asciiDocTable = new AsciiDocTable("Table of Attribute's constraints");
            asciiDocTable.SetHeaderRow(new List<string>() { "Name", "Constraint", "Type", "Expression of constraint" });

            if (at.IsConst)
            {
                asciiDocTable.AddRow(new List<string>() { at.Name, "Constant", "", at.Default.ToString() });
                constraintsExist = true;
            }

            if (at.IsStatic)
            {
                asciiDocTable.AddRow(new List<string>() { at.Name, "Fixed", "", at.Default.ToString() });
                constraintsExist = true;
            }

            if (at.Constraints.Count > 0)
            {
                constraintsExist = true;
                foreach (EA.AttributeConstraint ac in at.Constraints)
                {
                    string constrNotes = ac.Notes;
                    string constraint = "";
                    if (ac.Type == "OCL")
                    {
                        int pos1 = constrNotes.IndexOf("->");
                        int pos2 = constrNotes.IndexOf('(');
                        int pos3 = constrNotes.IndexOf(')');
                        constraint = constrNotes.Substring(pos1 + 1, pos3 - pos1);
                    }
                    else
                    {
                        if (ac.Type.Equals("INV") || ac.Type.Equals("Invariant"))
                        {
                            constraint = ">" + constrNotes;
                        }

                    }
                    if (constraint.Length > 0)
                    {
                        asciiDocTable.AddRow(new List<string>() { at.Name, ac.Name, ac.Type, constraint });
                    }
                }
            }
            if (constraintsExist)
            {
                sb.Append(asciiDocTable.GetTable("1, 2, 1, 3", 100));
                return sb.ToString() + LBR + LBR;
            }
            return "";
        }


        private string GetAsciiDocWithAttributesTable(EA.Element elem, string type)
        {
            StringBuilder sb = new StringBuilder();
            AsciiDocStyle asciiDocStyle = new AsciiDocStyle();

            sb.Append(asciiDocStyle.GetHeader(elem.Name + " " + type, 3));
            sb.Append(elem.Notes + LBR);
            string packName = repo.GetPackageByID(elem.PackageID).Name;
            sb.Append("The following table shows all attributes of " + packName + "::" + elem.Name + ": " + LBR);

            AsciiDocTable asciiDocTable = new AsciiDocTable("Table of Attributes of " + packName + "::" + elem.Name);
            asciiDocTable.SetHeaderRow(new List<string>() { "Mult.", "Attribute Name / Type", "Description" });
            foreach (EA.Attribute attr in elem.Attributes)
            {
                asciiDocTable.AddRow(new List<string>() { "[" + attr.LowerBound + ".." + attr.UpperBound + "]",
                                                          attr.Name + " /" + LBR + attr.Type,
                                                          attr.Notes});
            }
            sb.Append(asciiDocTable.GetTable("1, 2, 4", 100));
            return sb.ToString();
        }


        public bool IsClassWithoutStereotype(EA.Element elt)
        {
            if (elt.StereotypeEx == "") return true;

            if (elt.Type == "Enumeration") return false;

            foreach (string stereo in this.notPureClassStereotypes)
            {
                if (elt.StereotypeEx.Contains(stereo))
                {
                    return false;
                }
            }
            return true;
        }

        private string GetElementType(EA.Element el)
        {
            string ret = "";
            if ((el.StereotypeEx.Contains(CD.GetEnumStereotype())) || (el.Type == "Enumeration")) ret = "enumeration";
            if (el.StereotypeEx.Contains(CD.GetDatatypeStereotype())) ret = "datatype";
            if (el.StereotypeEx.Contains(CD.GetCompoundStereotype())) ret = "compound";
            if (el.StereotypeEx.Contains("datatype")) ret = "Datatype"; // legacy
            if (el.StereotypeEx.Contains(CD.GetPrimitiveStereotype())) ret = "primitive";
            return ret;
        }

        private void SelectStereotypedClasses(List<EA.Element> elements)
        {
            foreach (EA.Element elt in elements)
            {
                string elttype = GetElementType(elt);
                if (TypeList.ContainsKey(elttype))
                {
                    if (!TypeList[elttype].Contains(elt.ElementID)) TypeList[elttype].Add(elt.ElementID);
                }
            }
        }


    }
}
