﻿using CimSyntaxGen.Utilities;
using System;
using System.Collections.Generic;
using CimSyntaxGen.JsonMaster;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CimSyntaxGen.AvroGeneration;
using CimSyntaxGen.JsonGeneration;
using System.Web;
using System.Windows.Forms;
using System.Diagnostics.Eventing.Reader;
using EA;
using System.Text.Json.Nodes;

namespace CimSyntaxGen.AvroGeneration
{
    public class AvroMapper
    {
        public static string AVRO_FILE_EXTENSION = ".avsc";
        private EA.Repository repo;
        private Logger logger = new Logger(Main.LOG_FILE, Main.DO_LOG);
        private static string CONN_TYPE_ASSOCIATION = "Association";
        private static string CONN_TYPE_AGGREGATION = "Aggregation";
        private static string CONN_TYPE_GENERALIZATION = "Generalization";
        private static string CONN_DIR_SRC_DEST = "Source -> Destination";
        private static string CONN_DIR_DEST_SRC = "Destination -> Source";
        private HashSet<string> propertiesSet = new HashSet<string>();
        
        public AvroMapper(EA.Repository repo) 
        {
            this.repo = repo;
        } 

        public string CreateMapping(EA.Package entryPackage, string nameSpace)
        {
            MsgBox msgBox = new MsgBox();
            msgBox.ShowProcessing();

            List<int> elemIDList = GetElementIDsInPackage(entryPackage);
            AvroRecord envelopeAvro = new AvroRecord(entryPackage, nameSpace);

            List<EA.Element> rootElements = GetRootElements(entryPackage);
            if(rootElements.Count <= 0)
            {
                msgBox.Close();
                MessageBox.Show("Error: No root element defined!");
                return null;
            }
            foreach(EA.Element rootElement in rootElements)
            {
                AvroRecord rootAvro = new AvroRecord(rootElement, nameSpace);
                GetProperties(rootElement, rootAvro.GetFields(), nameSpace);
                // go thru connectors
                CreateLinkedProperties(rootElement, rootAvro, nameSpace, elemIDList);
                GetInheritedProperties(rootElement, rootAvro, nameSpace);
                // check if root element is an array, is so, create an array and
                // add the root element to it as item
                int[] minMax = GetRootMinMax(rootElement);
                if (minMax[0] != minMax[1]) // array 
                {
                    AvroArray<JsonBase> rootArrayAvro = new AvroArray<JsonBase>();
                    rootArrayAvro.AddItem(rootAvro);
                    AvroField rootFieldArray = new AvroField(rootElement.Name, rootArrayAvro, null, null);
                    envelopeAvro.AddFields(rootFieldArray);
                    //envelopeAvro.AddFields(rootArrayAvro);
                }
                else
                {
                    AvroField rootField = new AvroField(rootElement.Name, rootAvro, null, null);
                    envelopeAvro.AddFields(rootField);
                    //envelopeAvro.AddFields(rootAvro);
                }
            }
            msgBox.Close();
            return envelopeAvro.AsString();
        }

        private List<int> GetElementIDsInPackage(EA.Package pack)
        {
            List<int> elemIDList = new List<int>(); 
            foreach(EA.Element elem in pack.Elements) // note: contains() of EA API is not reliable
            {
                elemIDList.Add(elem.ElementID);
            }
            return elemIDList;
        }

        // Is recursively called from entryElem
        private void CreateLinkedProperties(EA.Element entryElem, AvroRecord entryAvro, string nameSpace, List<int> elemIDList)
        {
            foreach(EA.Connector conn in entryElem.Connectors)
            {
                if(conn.Type.Equals(CONN_TYPE_ASSOCIATION) || conn.Type.Equals(CONN_TYPE_AGGREGATION))
                {
                    EA.Element otherElem = null;
                    String roleName = "";
                    String cardinality = null;
                    if(conn.Type.Equals(CONN_TYPE_ASSOCIATION))
                    {
                        if ((entryElem.ElementID == conn.SupplierID) && (conn.Direction.Equals(CONN_DIR_DEST_SRC)))
                        {
                            otherElem = repo.GetElementByID(conn.ClientID);
                            roleName = conn.ClientEnd.Role;
                            cardinality = conn.ClientEnd.Cardinality;
                        }
                        else if ((entryElem.ElementID == conn.ClientID) && (conn.Direction.Equals(CONN_DIR_SRC_DEST)))
                        {
                            otherElem = repo.GetElementByID(conn.SupplierID);
                            roleName = conn.SupplierEnd.Role;
                            cardinality = conn.SupplierEnd.Cardinality;
                        }
                    } else // Aggregation or Composition
                    {
                        if ((entryElem.ElementID == conn.SupplierID) && 
                            conn.Direction.Equals(CONN_DIR_SRC_DEST) && 
                            conn.SupplierEnd.Aggregation > 0) // Aggr. = 1 ...Aggregation, Aggr. = 2 ...Composition
                        {
                            otherElem = repo.GetElementByID(conn.ClientID);
                            roleName = conn.ClientEnd.Role;
                            cardinality = conn.ClientEnd.Cardinality;
                        }
                    }

                    // we handle only elements of the package!
                    if(otherElem != null && elemIDList.Contains(otherElem.ElementID))
                    {
                        AvroField otherElemField = null;
                        if (otherElem.Type.Equals("Enumeration"))
                        {
                            AvroEnum otherElemAvro = GetEnumerationAvro(otherElem, nameSpace);
                            // check for cardinality, if array, then wrap it up in Avro Array
                            if (GetMaxCardinality(cardinality) != GetMinCardinality(cardinality))
                            {
                                AvroArray<AvroRecord> avroArray = new AvroArray<AvroRecord>(otherElemAvro);
                                otherElemField = new AvroField(roleName, avroArray, null, null);
                            }
                            else
                            {
                                otherElemField = new AvroField(roleName, otherElemAvro, null, null);
                            }
                        }
                        else
                        {
                            if(propertiesSet.Contains(otherElem.Name))
                            {
                                if (GetMaxCardinality(cardinality) != GetMinCardinality(cardinality))
                                {
                                    AvroArray<string> avroArray = new AvroArray<string>(AvroRecord.GetAvroCaseName(otherElem.Name));
                                    otherElemField = new AvroField(roleName, avroArray, null, null);
                                }
                                else
                                {
                                    otherElemField = new AvroField(roleName, otherElem.Name, null, null);
                                }
                            }
                            else
                            {
                                AvroRecord otherElemAvro = new AvroRecord(otherElem, nameSpace);
                                GetProperties(otherElem, otherElemAvro.GetFields(), nameSpace);
                                GetInheritedProperties(otherElem, otherElemAvro, nameSpace);
                                // recursive part
                                CreateLinkedProperties(otherElem, otherElemAvro, nameSpace, elemIDList);
                                // check for cardinality, if array, then wrap it up in Avro Array
                                if (GetMaxCardinality(cardinality) != GetMinCardinality(cardinality))
                                {
                                    AvroArray<AvroRecord> avroArray = new AvroArray<AvroRecord>(otherElemAvro);
                                    otherElemField = new AvroField(roleName, avroArray, null, null);
                                }
                                else
                                {
                                    otherElemField = new AvroField(roleName, otherElemAvro, null, null);
                                }
                            }
                        }
                        entryAvro.AddFields(otherElemField);
                        if(!propertiesSet.Contains(otherElem.Name)) propertiesSet.Add(otherElem.Name);
                    }
                }
            }
        }

        private void GetInheritedProperties(EA.Element entryElem, AvroRecord entryAvro, string nameSpace)
        {
            Stack<EA.Element> stack = new Stack<Element> ();
            foreach (EA.Connector conn in entryElem.Connectors)
            {
                if(conn.Type.Equals(CONN_TYPE_GENERALIZATION) && (conn.ClientID == entryElem.ElementID))
                {
                    Element parent = repo.GetElementByID(conn.SupplierID);
                    stack.Push(parent);
                }
            }
            while(stack.Count > 0)
            {
                Element parent = stack.Pop();
                GetParentProperties(parent, entryAvro.GetFields(), nameSpace);
                foreach (EA.Connector conn in parent.Connectors)
                {
                    if (conn.Type.Equals(CONN_TYPE_GENERALIZATION) && (conn.ClientID == parent.ElementID))
                    {
                        Element grandParent = repo.GetElementByID(conn.SupplierID);
                        stack.Push(grandParent);
                    }
                }
            }
        }

        private void GetParentProperties(EA.Element parent, JsonArray<JsonBase> fields, string nameSpace)
        {
            foreach (EA.Attribute attr in parent.Attributes)
            {
                if (attr.Visibility.Equals("Private")) continue;

                AvroField avroField;
                int lowerBound = -1;
                bool res = int.TryParse(attr.LowerBound, out lowerBound);

                if (attr.ClassifierID <= 0) // primitive data type
                {
                    string[] type = GetBasicType(attr.Type);
                    if (res && lowerBound == 0)
                    {
                        JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>();
                        jsonArray.Add(new AvroNull());
                        if (type[1] != null)
                        {
                            jsonArray.Add(AvroField.GetLogicalType(type[1], type[0]));
                        } else
                        {
                            jsonArray.Add(new JsonString(type[0]));
                        }
                        avroField = new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes);
                    }
                    else
                    {
                        if (type[1] != null)
                        {
                            avroField = new AvroField(attr.Name, type[1], type[0], attr.Default, attr.Notes);
                        }
                        else
                        {
                            avroField = new AvroField(attr.Name, type[0], attr.Default, attr.Notes);
                        }
                    }
                    fields.Add(avroField);
                }
                else
                {
                    EA.Element propElem = repo.GetElementByID(attr.ClassifierID);
                    JsonBase jsonPropElem = GetPropElementAvro(attr, propElem, nameSpace);
                    if (res && lowerBound == 0)
                    {
                        JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>("type");
                        jsonArray.Add(new AvroNull());
                        jsonArray.Add(jsonPropElem);
                        avroField = new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes);
                    }
                    else
                    {
                        avroField = new AvroField(attr.Name, jsonPropElem, attr.Default, attr.Notes);
                    }
                    fields.Add(avroField);
                }
            }
        }

        private void GetProperties(EA.Element elem, JsonArray<JsonBase> fields, string nameSpace)
        {
            foreach(EA.Attribute attr in elem.Attributes)
            {
                AvroField avroField;
                int lowerBound = -1;
                bool res = int.TryParse(attr.LowerBound, out lowerBound);

                if (attr.ClassifierID <= 0) // primitive data type
                {
                    string[] type = GetBasicType(attr.Type);
                    if(res && lowerBound == 0)
                    {
                        JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>();
                        jsonArray.Add(new AvroNull());
                        if (type[1] != null)
                        {
                            jsonArray.Add(AvroField.GetLogicalType(type[1], type[0]));
                        }
                        else
                        {
                            jsonArray.Add(new JsonString(type[0]));
                        }
                        avroField = new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes);
                    } else
                    {
                        if (type[1] != null)
                        {
                            avroField = new AvroField(attr.Name, type[1], type[0], attr.Default, attr.Notes);
                        }
                        else
                        {
                            avroField = new AvroField(attr.Name, type[0], attr.Default, attr.Notes);
                        }
                    }
                    fields.Add(avroField);
                } else
                {
                    EA.Element propElem = repo.GetElementByID(attr.ClassifierID);
                    if(propertiesSet.Contains(propElem.Name))
                    {
                        if (res && lowerBound == 0)
                        {
                            JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>("type");
                            jsonArray.Add(new AvroNull());
                            jsonArray.Add(new JsonString(AvroRecord.GetAvroCaseName(propElem.Name)));
                            avroField = new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes);
                        }
                        else
                        {
                            avroField = new AvroField(attr.Name, AvroRecord.GetAvroCaseName(propElem.Name), attr.Default, attr.Notes);
                        }
                    }
                    else
                    {
                        JsonBase jsonPropElem = GetPropElementAvro(attr, propElem, nameSpace);
                        if (res && lowerBound == 0)
                        {
                            JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>("type");
                            jsonArray.Add(new AvroNull());
                            jsonArray.Add(jsonPropElem);
                            avroField = new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes);
                        }
                        else
                        {
                            avroField = new AvroField(attr.Name, jsonPropElem, attr.Default, attr.Notes);
                        }
                    }
                    fields.Add(avroField);
                    if(!propertiesSet.Contains(propElem.Name)) propertiesSet.Add(propElem.Name);
                }
            }
        }

        private AvroEnum GetEnumerationAvro(EA.Element elem, string nameSpace)
        {
            return new AvroEnum(elem, nameSpace);
        }

        private JsonBase GetClassAvro(EA.Element elem, string nameSpace)
        {
            AvroRecord classAvro = new AvroRecord(elem, nameSpace);
            GetProperties(elem, classAvro.GetFields(), nameSpace);
            return classAvro;
        }

        private JsonBase GetPrimitive(EA.Element elem, string nameSpace)
        {
            AvroRecord primitiveAvro = new AvroRecord(elem, nameSpace);
            foreach(EA.Constraint constraint in elem.Constraints)
            {
                primitiveAvro.AddFields(new AvroField(constraint.Name, constraint.Notes, null, null));
            }
            return primitiveAvro;
        }

        private JsonBase GetCompoundAvro(EA.Element elem, string nameSpace)
        {
            AvroRecord compoundAvro = new AvroRecord(elem, nameSpace);
            GetProperties(elem, compoundAvro.GetFields(), nameSpace);
            return compoundAvro;
        }

        private JsonBase GetCIMDatatypeAvro(EA.Element elem, string nameSpace)
        {
            AvroRecord cimDatatypeAvro = new AvroRecord(elem, nameSpace);
            foreach (EA.Attribute attr in elem.Attributes)
            {
                JsonBase attrType = new JsonString("");
                string[] basicType = GetBasicType(attr.Type);
                if (basicType != null) {
                    if (basicType[1] != null)
                    {
                        attrType = AvroField.GetLogicalType(basicType[1], basicType[0]);
                    }
                    else
                    {
                        attrType = new JsonString(basicType[0]);
                    }
                }

                int lowerBound = -1;
                bool res = int.TryParse(attr.LowerBound, out lowerBound);
                if(res && lowerBound == 0)
                {
                    JsonArray<JsonBase> jsonArray = new JsonArray<JsonBase>();
                    jsonArray.Add(new AvroNull());
                    jsonArray.Add(attrType);
                    logger.Log("CIMDatatype array: \n" + jsonArray.AsString());
                    cimDatatypeAvro.AddFields(new AvroField(attr.Name, jsonArray, attr.Default, attr.Notes));
                } else
                {
                    AvroEnum stdCodelistEnum = null;
                    EA.Element classifier = null;
                    if (attr.Name.Equals("value"))
                    {
                        classifier = repo.GetElementByID(attr.ClassifierID);
                        if(classifier.Stereotype.ToLower().Equals("enumeration") && classifier.Name.ToLower().Contains("typelist"))
                        {
                            stdCodelistEnum = new AvroEnum("Standard" + classifier.Name.Replace("TypeList", "Type"));
                            stdCodelistEnum.Namespace = nameSpace;
                            stdCodelistEnum.SetDoc(classifier.Notes);
                            foreach(EA.Attribute classifierAttr in classifier.Attributes)
                            {
                                stdCodelistEnum.AddSymbol(classifierAttr.Default);
                            }
                        }
                    }
                    AvroField attrField;
                    if (stdCodelistEnum != null && classifier != null) {
                        AvroEnum localCodelistEnum = new AvroEnum("Local" + classifier.Name.Replace("TypeList", "Type"));
                        localCodelistEnum.Namespace = nameSpace;
                        localCodelistEnum.AddSymbol("A01");
                        localCodelistEnum.SetDoc("");
                        JsonArray<JsonBase> codelists = new JsonArray<JsonBase>("type");
                        codelists.Add(stdCodelistEnum);
                        codelists.Add(localCodelistEnum);
                        attrField = new AvroField(attr.Name, codelists, attr.Default, attr.Notes);
                    }
                    else 
                    {
                        attrField = new AvroField(attr.Name, attrType, attr.Default, attr.Notes);
                    }
                    cimDatatypeAvro.AddFields(attrField);
                }
            }
            return cimDatatypeAvro;
        }
        private JsonBase GetPropElementAvro(EA.Attribute attr, EA.Element elem, string nameSpace)
        {
            JsonBase propElem = null;
            if (elem.Type.Equals("Enumeration"))
            {
                propElem = GetEnumerationAvro(elem, nameSpace);
            }
            else if (elem.Type.Equals("Class"))
            {
                if (elem.Stereotype == null || elem.Stereotype.Equals(""))
                {
                    if (elem.Abstract.Equals("1")) return null;
                    else
                    {
                        propElem = GetClassAvro(elem, nameSpace);
                    }
                }
                else if (elem.Stereotype.Contains("enumeration") || elem.Stereotype.Contains("Enumeration"))
                {
                    propElem = GetEnumerationAvro(elem, nameSpace);
                }
                else if (elem.Stereotype.Contains("Compound"))
                {
                    propElem = GetCompoundAvro(elem, nameSpace);
                }
                else if (elem.Stereotype.Contains("CIMDatatype"))
                {
                    propElem = GetCIMDatatypeAvro(elem, nameSpace);
                }
                else if (elem.Stereotype.Contains("Primitive"))
                {
                    propElem = GetPrimitive(elem, nameSpace);
                }
                else
                {
                    logger.Log("Invalid type: " + elem.Name);
                    MessageBox.Show("Invalid type: " + elem.Name);
                }
            }
            //AvroRecord avroAttribute = new AvroRecord(attr, propElem, nameSpace);
            //return avroAttribute;
            return propElem;
        }

        private string[] GetBasicType(string name)
        {
            if (name.Equals("Boolean"))
            {
                return new string[] { AvroTypes.PRIM_BOOLEAN, null };
            }
            else if (name.Equals("Decimal"))
            {
                return new string[] { AvroTypes.LOG_DECIMAL, "int"};
            }
            else if (name.Equals("Float"))
            {
                return new string[] { AvroTypes.PRIM_FLOAT, null };
            }
            else if (name.Equals("Integer"))
            {
                return new string[] { AvroTypes.PRIM_INT, null };
            }
            else if (name.Equals("Double"))
            {
                return new string[] { AvroTypes.PRIM_DOUBLE, null };
            }
            else if (name.Equals("String"))
            {
                return new string[] { AvroTypes.PRIM_STRING, null };
            }
            else if (name.Equals("DateTime"))
            {
                return new string[] { AvroTypes.LOG_TIMESTAMP_MILLIS, "long"};
            }
            else if (name.Equals("Date"))
            {
                return new string[] { AvroTypes.LOG_DATE, "int"};
            }
            else if (name.Equals("Duration"))
            {
                return new string[] { AvroTypes.LOG_DURATION, "fixed"};
            }
            else if (name.Equals("Time"))
            {
                return new string[] { AvroTypes.LOG_TIME_MILLIS, "int"};
            }
            else if (name.Equals("MonthDay"))
            {
                return new string[] { AvroTypes.PRIM_STRING, null};
            }
            return new string[] { AvroTypes.PRIM_STRING, null };
        }

        private int[] GetRootMinMax(EA.Element rootElem)
        {
            int min = -1;
            int max = -1;
            foreach (EA.Connector conn in rootElem.Connectors)
            {
                if(conn.Type.Equals("Composition") || conn.Type.Equals("Aggregation"))
                {
                    return new int[] { 1, 1 };
                }
                EA.ConnectorEnd connEnd = (conn.ClientID == rootElem.ElementID)? conn.ClientEnd : conn.SupplierEnd;
                string cardinality = connEnd.Cardinality;
                if (cardinality != null && cardinality.Length > 0)
                {
                    if (!cardinality.StartsWith("1"))
                    {
                        string lowerBound = cardinality.Substring(0, 1);
                        if (lowerBound.Equals('*')) min = 0;
                        else
                        {
                            int tmp = -1;
                            try { tmp = Int32.Parse(lowerBound); }
                            catch (Exception e)
                            {
                                ErrorCodes.ShowException(ErrorCodes.ERROR_044, "At " + rootElem.Name, e);
                            }
                            if (min < 0) min = tmp;
                            else if (min > tmp) min = tmp;
                        }
                    }
                    if ((cardinality.Length >= 2) && !cardinality.EndsWith("1"))
                    {
                        string upperBound = cardinality.Substring(cardinality.Length - 2, 1);
                        if (upperBound.Equals('*')) max = -1;
                        else
                        {
                            int tmp = -1;
                            try { tmp = Int32.Parse(upperBound); } 
                            catch (Exception e) 
                            {
                                ErrorCodes.ShowException(ErrorCodes.ERROR_045, "At " + rootElem.Name, e);
                            }
                            if (max < tmp) max = tmp; 
                        }
                    }
                }
            }
            return new int[] { min, max };
        }

        private List<EA.Element> GetRootElements(EA.Package entryPackage)
        {
            List<EA.Element> rootElements = new List<EA.Element>();
            foreach(EA.Element elem in entryPackage.Elements)
            {
                if(elem.IsActive) // 'isActive' attribute indicates root elements
                {
                    rootElements.Add(elem);
                }
            }
            return rootElements;
        }

        private int GetMinCardinality(string cardinality)
        {
            // No cardinality specified
            if (cardinality == null || cardinality.Equals("")) return 1;

            // there's a minimum specified, or there's just one character
            string ch = cardinality.Substring(0, 1);
            if ("0123456789".Contains(ch)) return Convert.ToInt16(ch);

            return 1; // Default
        }

        // Returns -1 for infinity or arbitrary many
        private int GetMaxCardinality(string cardinality)
        {
            // no cardinality specified
            if (cardinality == null || cardinality.Equals("")) return 1;

            // There is one max value specified or there's is just one character
            int len = cardinality.Length;
            if (len >= 1)
            {
                string lastChar = cardinality.Substring(len - 1, 1);
                if ("123456789".Contains(lastChar)) return Convert.ToInt16(lastChar);
                if (lastChar.Equals("*")) return -1;
            }
            return 1; // Default value
        }


    }
}
