﻿/////////////////////////////////////////////////////////////////////////////////////////
// Author: Alexander Balka
// File: ConfigurationManager.cs
/////////////////////////////////////////////////////////////////////////////////////////


using CimSyntaxGen.Utilities;
using EA;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Remoting.Contexts;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Schema;

namespace CimSyntaxGen.Configuration
{
    public class ConfigurationManager
    {
        public static readonly string CONFIG_PACKAGE = ".Configuration";
        public static readonly string CONFIG_ELEMENT = "CimSyntaxGen";
        public static readonly string CONFIG_TAGGGEDVALUE = "configuration";
        private EA.Package theModel = null;
        private readonly EA.Repository repo = null;
        private static ConfigurationManager configurationManager = null;
        private static string customConfiguration = null;
        public static readonly string CHECKED = "Checked";
        public static readonly string UNCHECKED = "Unchecked";
        public static string PARAM_CONFIGURATION = "configuration";
        public static string PARAM_APPSETTINGS = "appSettings";
        public static string PARAM_DATAPROFILE = "dataProfile";
        public static string[] validXmlParameter = { PARAM_CONFIGURATION, "profdata", "shacldata", "profstereo", "jsondata", "avrodata", "asciidocdata" };
        private static Logger logger = new Logger(Main.LOG_FILE, Main.DO_LOG);

        private ConfigurationManager(EA.Repository repository)
        {
            repo = repository;
        }

        public static ConfigurationManager GetConfigurationManager(EA.Repository repo) // access to singleton
        {
            if(configurationManager == null)
            {
                configurationManager = new ConfigurationManager(repo);
                configurationManager.SetModel();
                configurationManager.StartUp();
            }
            return configurationManager;
        }

        public void DefineStereotypes()
        {
            EA.Stereotype stereotype;
            string name = "Enumeration";
            if (!ExistsStereotype(name) && !ExistsStereotype(name.ToLower()))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Class");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }

            name = "Compound";
            if (!ExistsStereotype(name))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Class");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }

            name = "CIMDatatype";
            if (!ExistsStereotype(name))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Class");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }

            name = "Primitive";
            if (!ExistsStereotype(name))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Class");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }

            name = "IsBasedOn";
            if (!ExistsStereotype(name))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Dependency");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }

            name = "Description";
            if (!ExistsStereotype(name) && !ExistsStereotype(name.ToLower()))
            {
                stereotype = (EA.Stereotype)repo.Stereotypes.AddNew(name, "Class");
                stereotype.Notes = "CIM Profile Stereotype";
                stereotype.Update();
                repo.Stereotypes.Refresh();
            }
        }

        private bool ExistsStereotype(string name)
        {
            foreach(EA.Stereotype stereotype in repo.Stereotypes)
            {
                if(stereotype.Name.Equals(name)) return true;
            }
            return false;
        }

        private void SetModel()
        {
            EA.Collection models = repo.Models;
            try
            {
                if (models.Count > 1) 
                {   // There are several. Let's try to find the one with the configuration  package. 
                    // If there is no configuration package, then take the first model.
                    foreach (EA.Package model in models)
                    {
                        EA.Package pack = GetConfigPackage(model);
                        if (pack != null)
                        {
                            theModel = (EA.Package)model;
                            return;
                        }
                    }
                }
                theModel = (EA.Package)models.GetAt(0);
            }
            catch (Exception e)
            {
                MessageBox.Show(ErrorCodes.ERROR_005 + ": " + ErrorCodes.ERROR_005[1] + "\n" + e.Message);
            }
        }

        public EA.Package TheModel { get { return theModel; } }

        public string StartUp()
        {
            try
            {
                // check whether there's a custom configuration 
                string configXml = this.LoadCustomConfiguration();
                if (configXml != null)
                {
                    customConfiguration = configXml;
                } else
                {
                    // there's no custom configuration, hence take a copy of the default:
                    this.LoadDefaultConfiguration();
                }
            } 
            catch(Exception e)
            {
                MessageBox.Show(ErrorCodes.ERROR_004[0] + ": " + ErrorCodes.ERROR_004[1] + "\n" + e.StackTrace);
            }
            return customConfiguration;
        }

        public string GetCustomConfiguration()
        {
            //return customConfiguration;
            string configStr = LoadCustomConfiguration();
            if(configStr == null) 
            {
                configStr = LoadDefaultConfiguration();
            }
            return configStr;
        }

        public Stream GetCustomConfigurationAsStream()
        {
            byte[] byteArray = Encoding.ASCII.GetBytes(customConfiguration);
            return new MemoryStream(byteArray);
        }

        public string LoadCustomConfiguration()
        {
            this.theModel.Packages.Refresh(); // in case of someone deleted the custom configuration manually
            EA.Package configPkg = GetConfigPackage(theModel);
            if (configPkg != null)
            {
                EA.Element configElem = GetExistingConfigElement(configPkg);
                if (configElem != null)
                {
                    EA.TaggedValue tv = GetExistingConfigTaggedValue(configElem) ?? CreateConfigTaggedValue(configElem);
                    string configXml = tv.Notes;
                    if (configXml != null && configXml.Length > 0)
                    {
                        customConfiguration = configXml.Trim(); 
                        return customConfiguration;
                    }
                }
            }
            // there's no custom configuration
            return null;
        }

        public string LoadDefaultConfiguration()
        {
            Assembly myAssembly = Assembly.GetExecutingAssembly();
            Stream stream = myAssembly.GetManifestResourceStream("CimSyntaxGen.Resources.CimSyntaxgen-Config.xml");
            StreamReader reader = new StreamReader(stream);
            customConfiguration = reader.ReadToEnd();
            return customConfiguration;
        }


        public EA.Element GetConfigElement()
        {
            EA.Package configPkg = GetConfigPackage(theModel) ?? this.CreateConfigPackage(theModel);
            EA.Element configElem = GetExistingConfigElement(configPkg) ?? CreateConfigElement(configPkg);
            return configElem;
        }

        private EA.Package CreateConfigPackage(EA.Package theModel)
        {
            EA.Package pkg = (EA.Package)theModel.Packages.AddNew(CONFIG_PACKAGE, "Nothing");
            pkg.Update();
            theModel.Packages.Refresh();
            return pkg;
        }

        private EA.Element CreateConfigElement(EA.Package pkg)
        {
            EA.Element elem = (EA.Element)pkg.Elements.AddNew(CONFIG_ELEMENT, "Class");
            elem.Update();
            pkg.Elements.Refresh();
            return elem;
        }

        public EA.TaggedValue CreateConfigTaggedValue(EA.Element elem)
        {
            EA.TaggedValue tv = (EA.TaggedValue)elem.TaggedValues.AddNew(CONFIG_TAGGGEDVALUE, "");
            tv.Value = "<memo>";
            tv.Notes = "";
            tv.Update();
            elem.TaggedValues.Refresh();
            return tv;
        }

        public EA.TaggedValue GetExistingConfigTaggedValue(EA.Element elem)
        {
            EA.Collection tags = elem.TaggedValues;
            for (short i = 0; i < tags.Count; i++)
            {
                EA.TaggedValue tv = (EA.TaggedValue)tags.GetAt(i);
                if (tv.Name.Equals(CONFIG_TAGGGEDVALUE))
                {
                    return tv;
                }
            }
            return null;
        }

        private EA.Package GetConfigPackage(EA.Package theModel)
        {
            EA.Collection pkgColl = theModel.Packages;
            for (short i = 0; i < pkgColl.Count; i++) // foreach is unstable here!
            {
                EA.Package pack = (EA.Package)pkgColl.GetAt(i);
                if (pack.Name.Equals(CONFIG_PACKAGE))
                {
                    return pack;
                }
            }
            return null;
        }

        private EA.Element GetExistingConfigElement(EA.Package pkg)
        {
            foreach (EA.Element element in pkg.Elements)
            {
                if (element.Name.Equals(CONFIG_ELEMENT))
                {
                    return element;
                }
            }
            return null;
        }

        public void ExportConfiguration()
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog
            {
                InitialDirectory = @"C:\",
                DefaultExt = "xml",
                Filter = @"XML (*.xml)|*.xml",
                FilterIndex = 1,
                Title = @"Save Configuration File"
            };
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                if (saveFileDialog.FileName != "")
                {

                    Stream fileStream;
                    if ((fileStream = saveFileDialog.OpenFile()) != null)
                    {
                        byte[] info = new UTF8Encoding(true).GetBytes(ConfigurationManager.GetConfigurationManager(repo).GetCustomConfiguration());
                        fileStream.Write(info, 0, info.Length);
                        fileStream.Close();
                    }
                }
                saveFileDialog.Dispose();                    
            } 
            else // ABA20230504
            {
                saveFileDialog.Dispose ();
            }
        }

        public void StoreImportedConfiguration(string configContent)
        {
            EA.Element configElem = GetConfigElement();
            EA.TaggedValue tv = GetExistingConfigTaggedValue(configElem);
            if (tv == null)
            {
                tv = CreateConfigTaggedValue(configElem);
            }
            tv.Value = "<memo>";
            tv.Notes = configContent;
            tv.Update();
            configElem.TaggedValues.Refresh();
            string customConfig = LoadCustomConfiguration();
            if (customConfig == null)
            {
                MessageBox.Show(ErrorCodes.ERROR_006[0] + ": " + ErrorCodes.ERROR_006[1]);
            }
            MatchConfiguration();
        }


        public void ImportConfiguration()
        {
            ImportConfigurationDialog importConfigDialog = new ImportConfigurationDialog(repo);
            importConfigDialog.ShowDialog();
            importConfigDialog.Dispose();
        }

        public string ValidateConfigXML(string xmlString, string xsdSting)
        {
            XmlDocument xml = new XmlDocument();
            xml.LoadXml(xmlString);
            XmlReader xmlReader = XmlReader.Create(new StringReader(xsdSting));
            xml.Schemas.Add(null, xmlReader);

            try
            {
                xml.Validate(null);
            }
            catch (XmlSchemaValidationException ex)
            {
                return ex.Message;
            }
            return null;
        }

        public static string AsString(XmlDocument xmlDoc)
        {
            using (StringWriter sw = new StringWriter())
            {
                using (XmlTextWriter tx = new XmlTextWriter(sw))
                {
                    tx.Formatting = Formatting.Indented;
                    xmlDoc.WriteContentTo(tx);
                    tx.Flush();
                    string strXmlText = sw.ToString();
                    return strXmlText;
                }
            }
        }

        public void UpdateConfiguration(XmlDocument xmlDoc)
        {
            this.theModel.Packages.Refresh();
            ConfigurationManager configManager = ConfigurationManager.GetConfigurationManager(repo);
            EA.Element configElem = configManager.GetConfigElement();
            EA.TaggedValue tv = configManager.GetExistingConfigTaggedValue(configElem) ?? configManager.CreateConfigTaggedValue(configElem);
            tv.Value = "<memo>";
            tv.Notes = AsString(xmlDoc);
            tv.Update();
            configElem.TaggedValues.Refresh();
            configManager.LoadCustomConfiguration();
        }
        
        public void MatchConfiguration()
        {
            string defaultConfigXMLString = this.LoadDefaultConfiguration();
            string customConfigXMLString = this.GetCustomConfiguration();

            Dictionary<string, Dictionary<string, string>> defConfigDict  = LoadXMLString(defaultConfigXMLString);
            Dictionary<string, Dictionary<string, string>> custConfigDict = LoadXMLString(customConfigXMLString);

            Dictionary<string, Dictionary<string, string>> missingElems = new Dictionary<string, Dictionary<string, string>>();
            foreach (KeyValuePair<string, Dictionary<string, string>> kvPair in defConfigDict)
            {
                string key = kvPair.Key;
                Dictionary<string, string> attributes = kvPair.Value;
                if(!custConfigDict.ContainsKey(key))
                {
                    missingElems.Add(key, attributes);
                }
            }
            
            if(missingElems.Count > 0)
            {
                foreach (KeyValuePair<string, Dictionary<string, string>> kvP in missingElems)
                {
                    custConfigDict.Add(kvP.Key, kvP.Value);
                }
                string xmlString = CreateXML(custConfigDict);
                StoreImportedConfiguration(xmlString);
            }
        }

        private string GetPrefix(string nodeText)
        {
            int pos = nodeText.IndexOf('.');
            if (pos == -1) return null;
            return nodeText.Substring(0, pos);
        }


        private string CreateXML(Dictionary<string, Dictionary<string, string>> configDictionary)
        {
            XmlDocument doc = new XmlDocument();
            XmlDeclaration xmlDeclaration = doc.CreateXmlDeclaration("1.0", "UTF-8", null);
            XmlElement root = doc.DocumentElement;
            doc.InsertBefore(xmlDeclaration, root);

            XmlElement elemAddIn = doc.CreateElement("add-in");
            XmlElement elemAppSettings = doc.CreateElement(PARAM_APPSETTINGS);
            XmlElement elemDataProfile = doc.CreateElement(PARAM_DATAPROFILE);
            foreach (KeyValuePair<string, Dictionary<string, string>> keyVal in configDictionary)
            {
                string keyAppSettingsPrefix = PARAM_APPSETTINGS + "." + PARAM_CONFIGURATION;
                string keyDataProfilePrefix = PARAM_DATAPROFILE + ".";
                if (keyVal.Key.StartsWith(keyAppSettingsPrefix))
                {
                    AppendElem(doc, elemAppSettings, ConfigurationManager.PARAM_CONFIGURATION, keyVal.Value);
                }
                else if(keyVal.Key.StartsWith(keyDataProfilePrefix))
                {
                    string keyOfDataProfile = keyVal.Key.Replace(PARAM_DATAPROFILE + ".", "");
                    AppendElem(doc, elemDataProfile, GetPrefix(keyOfDataProfile), keyVal.Value);
                }
            }
            elemAddIn.AppendChild(elemAppSettings);
            elemAddIn.AppendChild(elemDataProfile);
            doc.AppendChild(elemAddIn);
            StringWriterWithEncoding sw = new StringWriterWithEncoding(new StringBuilder(""), Encoding.UTF8);
            doc.Save(sw);
            return sw.ToString();
        }

        private void AppendElem(XmlDocument doc, XmlElement parent, string elemName, Dictionary<string, string> attrs)
        {
            try
            {
                XmlElement elem = doc.CreateElement(elemName);
                Dictionary<string, string> values = attrs;
                foreach (KeyValuePair<string, string> keyVal2 in values)
                {
                    XmlAttribute attribute = doc.CreateAttribute(keyVal2.Key);
                    attribute.Value = keyVal2.Value;
                    elem.SetAttributeNode(attribute);
                }
                parent.AppendChild(elem);
            }
            catch (Exception e)
            {
                MessageBox.Show("elemName = " + elemName + "\n" + e.StackTrace);
            }
        }



        private Dictionary<string, Dictionary<string, string>> LoadXMLString(string configXMLString)
        {
            Dictionary<string, Dictionary<string, string>> valueDictionary = new Dictionary<string, Dictionary<string, string>>();
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.LoadXml(configXMLString);
            string name = xmlDoc.DocumentElement.Name;
            foreach (XmlNode xmlNode in xmlDoc.DocumentElement.ChildNodes)
            {
                if (xmlNode.NodeType == XmlNodeType.Element)
                {
                    string nodeName = xmlNode.Name;
                    foreach (XmlNode xmlNode2 in xmlNode.ChildNodes)
                    {
                        if (ConfigurationManager.validXmlParameter.Contains(xmlNode2.Name))
                        {
                            string prefix = xmlNode2.Name;
                            string attrName = "";
                            Dictionary<string, string> attrValueDictionary = new Dictionary<string, string>();
                            foreach (XmlAttribute xmlAttr in xmlNode2.Attributes)
                            {
                                if (xmlAttr.Name.Equals("name"))
                                {
                                    attrName = xmlAttr.Value;
                                }
                                attrValueDictionary.Add(xmlAttr.Name, xmlAttr.Value);
                            }
                            string node2 = nodeName + "." + prefix + "." + attrName;
                            valueDictionary.Add(node2, attrValueDictionary);
                        }
                    }
                }
            }
            return valueDictionary;
        }

    }
}
