﻿using CimSyntaxGen.Utilities;
using CimSyntaxGen.JsonMaster;


using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using System.Diagnostics.Eventing.Reader;
using System.Collections;
using System.Runtime.Remoting.Lifetime;

namespace CimSyntaxGen.JsonGeneration
{
    public class JsonMapper
    {
        public static string SCHEMA_TYPE_WG16 = "wg16";
        public static string SCHEMA_TYPE_WG19 = "wg19";
        public static string SCHEMA_DEF_ID = "id";
        public static string SCHEMA_DEF_TITLE = "title";
        public static string SCHEMA_ELEM_ID = "$id";
        public static string SCHEMA_ELEM_REF = "$ref";
        public static string SCHEMA_ELEM_DEFS_OTHERS = "$defs";
        public static string SCHEMA_ELEM_DEFS_DRAFT07 = "definitions";
        public static string SCHEMA_DESC = "description";
        public static string SCHEMA_TYPE = "type";
        public static string SCHEMA_MIN = "minimum";
        public static string SCHEMA_MAX = "maximum";
        public static string SCHEMA_MODEL_REF = "modelReference";
        public static string SCHEMA_ADD_PROPS = "additionalProperties";
        public static string CONFIG_ROOT_URI = "RootModelURI";
        public static string MIN_ITEMS = "minItems";
        public static string MAX_ITEMS = "maxItems";
        public static string SCHEMA_ELEM_DEFS;
        public static string SCHEMA_ELEM_PROPERTIES = "properties";
        public static string SCHEMA_ELEM_ALLOF = "allOf";
        public static string SCHEMA_ELEM_ONEOF = "oneOf";
        public static string SCHEMA_ELEM_REQUIRED = "required";
        public static string SCHEMA_ELEM_ANYOF = "anyOf";
        public static string SCHEMA_ELEM_NOT = "not";
        public static string STEREOTYPE_UNION = "union";

        private readonly int OPTIONAL = 1;
        private readonly int REQUIRED = 0;

        private EA.Repository repo;
        private EA.Package profilePackage;
        private List<long[]> xorConnGroups;
        private List<JsonMaster.XOROwner> xorOwnerList;
        private XMLParser xmlParser;
        private string rootModelURI;
        private JsonPropertiesMapper jpMapper;
        private EA.Package codelistPackage = null;

        private Logger logger = new Logger(Main.LOG_FILE, Main.DO_LOG);

        public JsonMapper()
        {
        }

        public EA.Package CodelistPackage { get { return codelistPackage; } }

        public string GetDefs()
        {
            string schemaVersion = xmlParser.GetXmlValueJsonData("JsonSchemaVersion");
            if (schemaVersion != null && schemaVersion.ToLower().Contains("draft-07"))
            {
                return SCHEMA_ELEM_DEFS_DRAFT07;
            }
            return SCHEMA_ELEM_DEFS_OTHERS;
        }

        private void LogOwners(List<JsonMaster.XOROwner> xorOwners)
        {
            if (!logger.On()) return;
            foreach(XOROwner xorOwner in xorOwners)
            {
                logger.Log("owner: " + xorOwner.GetOwner().Name);
                logger.LogArray(xorOwner.GetConnList().ToArray());
            }
        }

        // schemaType: wg16 or others
        public string CreateMapping(EA.Repository repo, EA.Package profilePackage, string schemaType, bool exportCodelist)
        {
            MsgBox msgBox = new MsgBox();
            msgBox.ShowProcessing();
            this.repo = repo;
            this.xmlParser = new XMLParser(repo);
            this.profilePackage = profilePackage;
            xorConnGroups = JsonGenerationUtilities.GetXORConns(this.profilePackage);
            xorOwnerList = JsonGenerationUtilities.GetXOROwners(repo, profilePackage, xorConnGroups);
            this.rootModelURI = xmlParser.GetXmlValueProfData(CONFIG_ROOT_URI);
            this.jpMapper = new JsonPropertiesMapper(this.rootModelURI, this.repo, schemaType, xmlParser);
            SCHEMA_ELEM_DEFS = GetDefs();

            // HEADER ////////////////////////////////////////////////
            JsonKeyValue<string> jKVId;
            if (schemaType.Equals(SCHEMA_TYPE_WG16))
            { 
                jKVId = new JsonKeyValue<string>(SCHEMA_ELEM_ID, xmlParser.GetXmlValueJsonData("JsonSchemaID"));
            }
            else
            { 
                jKVId = new JsonKeyValue<string>(SCHEMA_ELEM_ID, xmlParser.GetXmlValueJsonData("JsonNameSpace").Trim('#') + ".schema.json");
            }
            JsonKeyValue<string> jKVSchema         = new JsonKeyValue<string>("$schema", xmlParser.GetXmlValueJsonData("JsonSchemaVersion"));
            string envelopName = xmlParser.GetXmlValueProfData("EnvelopName");            
            DateTime now = DateTime.Now; //2023-03-10 09:59:27.074276
            string formattedDateTime = now.ToString("yyyy-MM-dd HH:mm:ss.ffffffK");
            JsonKeyValue<string> jKVTitle          = new JsonKeyValue<string>("title", envelopName + "  profile " + "generated " + formattedDateTime);
            JsonKeyValue<string> jKVDescription    = new JsonKeyValue<string>(SCHEMA_DESC, JsonGenerationUtilities.ReplaceLFCR(profilePackage.Notes));
            JsonKeyValue<string> jKVNamespace      = new JsonKeyValue<string>("namespace", xmlParser.GetXmlValueJsonData("JsonNameSpace"));
            JsonKeyValue<string> jKVType           = new JsonKeyValue<string>(SCHEMA_TYPE, "object");
            JsonKeyValue<bool>   jKVAddProperties  = new JsonKeyValue<bool>(SCHEMA_ADD_PROPS, false);

            JsonGenerationUtilities jGUtil = new JsonGenerationUtilities();

            // get all elements and fill root elements in list rootElements
            List<EA.Element> rootElements = new List<EA.Element>();
            List<string> classifierAttrNameList = new List<string>();
            List<EA.Element> allElemsExclRoots = jGUtil.GetAllElements(repo, profilePackage, rootElements, classifierAttrNameList);

            // envelope ////////////////////////////////////////////////////
            JsonObject<JsonBase> jEnvObj = new JsonObject<JsonBase>(envelopName);
            JsonKeyValue<string> jRef = new JsonKeyValue<string>(SCHEMA_ELEM_REF, "#");
            jEnvObj.Add(jRef);

            // linked elements /////////////////////////////////////////////
            List<EA.Element> linkedAttrs = jGUtil.GetAllLinkedAttributes(repo, profilePackage, classifierAttrNameList);
            allElemsExclRoots.AddRange(linkedAttrs);

            // Json Schema ///////////////////////////////////////////////////////////
            JsonObject<JsonBase> jMapping = new JsonObject<JsonBase>("");
            jMapping.Add(jKVId);
            jMapping.Add(jKVSchema);
            jMapping.Add(jKVTitle);
            jMapping.Add(jKVDescription);
            jMapping.Add(jKVNamespace);
            jMapping.Add(jKVType);
            jMapping.Add(jKVAddProperties);

            // root elements
            List<JsonBase> jRootObjs = GetRootElements(rootElements);
            foreach (JsonBase rootProp in jRootObjs)
            {
                jMapping.Add(rootProp);
            }

            JsonObject<JsonBase> jDefinitions = new JsonObject<JsonBase>(SCHEMA_ELEM_DEFS);
            jDefinitions.Add(jEnvObj);

            List<JsonBase> rootSubSchemas = GetAllSubSchemas(rootElements);

            foreach(JsonBase rootSubSchema in rootSubSchemas)
            {
                jDefinitions.Add(rootSubSchema);
            }
            // add others
            List<JsonBase> subSchemas = new List<JsonBase>();
            try
            {
                subSchemas = GetAllSubSchemas(allElemsExclRoots);
            } catch(Exception ex)
            {
                MessageBox.Show(ex.Message + ":\n" + ex.StackTrace);
                msgBox.Close();
                return "";
            }
            //logger.Log("Count of subSchemas: " + subSchemas.Count);
            if(subSchemas != null)
            {
                foreach (JsonBase subSchema in subSchemas)
                {
                    jDefinitions.Add(subSchema);
                }
            }

            jMapping.Add(jDefinitions);
            msgBox.Close();
            return jMapping.AsString();
        }

        public string GenerateJsonCodelist(EA.Package selectedPkg, EA.Repository repo, string codelistSchemaId, string localCodelistFileName, string modelRefUri, string outputDir)
        {
            MsgBox msgBox = new MsgBox();
            msgBox.ShowProcessing();
            this.repo = repo;
            try
            {
                this.xmlParser = new XMLParser(repo);
                SCHEMA_ELEM_DEFS = GetDefs();

                // VERSION DATA //////////////////////////////////////////
                CodelistVersion codelistVersion = GetCodelistVersion(selectedPkg);
                if (codelistVersion == null)
                {
                    msgBox.Close();
                    MessageBox.Show("Missing 'Version' class in package!\nGeneration not possible.");
                    return null;
                }

                // HEADER ////////////////////////////////////////////////
                JsonObject<JsonBase> jMapping = new JsonObject<JsonBase>("");
                JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_ELEM_ID, codelistSchemaId);
                JsonKeyValue<string> jKVSchema = new JsonKeyValue<string>("$schema", xmlParser.GetXmlValueJsonData("JsonSchemaVersion"));
                JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>("title", "ENTSO - E Code Lists - " + codelistVersion.Version + " - " + codelistVersion.Release + " - release date = " + codelistVersion.ReleaseDate);
                JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, codelistVersion.Title);
                JsonKeyValue<string> jKVNamespace = new JsonKeyValue<string>("namespace", xmlParser.GetXmlValueProfData("URICodelist"));
                jMapping.Add(jKVId);
                jMapping.Add(jKVSchema);
                jMapping.Add(jKVTitle);
                jMapping.Add(jKVDescription);
                jMapping.Add(jKVNamespace);

                // codelists ///////////////////////////////////////////////////
                JsonObject<JsonBase> jCodelists = new JsonObject<JsonBase>(SCHEMA_ELEM_DEFS);
                JsonObject<JsonBase> jEnvObj = new JsonObject<JsonBase>("CodeLists");
                JsonKeyValue<string> jRef = new JsonKeyValue<string>(SCHEMA_ELEM_REF, "#");
                jEnvObj.Add(jRef);
                jCodelists.Add(jEnvObj);

                JsonPropertiesMapper jsonPropertiesMapper = new JsonPropertiesMapper(repo);
                bool codelistsExist = false;
                foreach (EA.Element elem in selectedPkg.Elements)
                {
                    if ((elem.Type.ToLower().Equals("enumeration") || elem.Stereotype.ToLower().Equals("enumeration")) && elem.Name.ToLower().Contains("typelist"))
                    {
                        if (!codelistsExist) codelistsExist = true;
                        JsonObject<JsonBase> codeListMapped = jsonPropertiesMapper.MapCodelist(elem, localCodelistFileName, modelRefUri, outputDir);
                        jCodelists.Add(codeListMapped);
                        JsonObject<JsonBase> stdCodelistMapped = jsonPropertiesMapper.MapStandardCodelist(elem);
                        jCodelists.Add(stdCodelistMapped);
                    }
                }
                if (!codelistsExist)
                {
                    msgBox.Close();
                    MessageBox.Show("No Codelist in selected package (no 'enumeration' stereotype and \nno 'typelist' in the element's name)!\nGeneration not possible.");
                    return null;
                }
                jMapping.Add(jCodelists);
                return jMapping.AsString();
            }
            finally
            {
                msgBox.Close();
            }
        }

        public string GenerateJsonCodelistTemplate(EA.Package selectedPkg, EA.Repository repo, string fullFileName)
        {
            MsgBox msgBox = new MsgBox();
            msgBox.ShowProcessing();
            this.repo = repo;
            try
            {
                this.xmlParser = new XMLParser(repo);
                SCHEMA_ELEM_DEFS = GetDefs();

                // HEADER ////////////////////////////////////////////////
                JsonObject<JsonBase> jMapping = new JsonObject<JsonBase>("");
                JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_ELEM_ID, @"file:///" + fullFileName.Replace("\\", "/"));
                JsonKeyValue<string> jKVSchema = new JsonKeyValue<string>("$schema", xmlParser.GetXmlValueJsonData("JsonSchemaVersion"));
                JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>("title", "");
                JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, "");
                JsonKeyValue<string> jKVNamespace = new JsonKeyValue<string>("namespace", "");
                jMapping.Add(jKVId);
                jMapping.Add(jKVSchema);
                jMapping.Add(jKVTitle);
                jMapping.Add(jKVDescription);
                jMapping.Add(jKVNamespace);

                // codelists ///////////////////////////////////////////////////
                JsonObject<JsonBase> jCodelists = new JsonObject<JsonBase>(SCHEMA_ELEM_DEFS);
                JsonObject<JsonBase> jEnvObj = new JsonObject<JsonBase>("CodeLists");
                JsonKeyValue<string> jRef = new JsonKeyValue<string>(SCHEMA_ELEM_REF, "#");
                jEnvObj.Add(jRef);
                jCodelists.Add(jEnvObj);
                JsonPropertiesMapper jsonPropertiesMapper = new JsonPropertiesMapper(repo);
                bool codelistsExist = false;
                foreach (EA.Element elem in selectedPkg.Elements)
                {
                    if ((elem.Type.ToLower().Equals("enumeration") || elem.Stereotype.ToLower().Equals("enumeration")) && elem.Name.ToLower().Contains("typelist"))
                    {
                        if (!codelistsExist) codelistsExist = true;
                        JsonObject<JsonBase> localCodeListMapped = jsonPropertiesMapper.MapLocalCodelist(elem);
                        jCodelists.Add(localCodeListMapped);
                    }
                }
                if (!codelistsExist)
                {
                    msgBox.Close();
                    MessageBox.Show("No Codelist in selected package (no 'enumeration' stereotype and \nno 'typelist' in the element's name)!\nGeneration not possible.");
                    return null;
                }
                jMapping.Add(jCodelists);
                return jMapping.AsString();
            }
            finally
            {
                msgBox.Close();
            }
        }

        private CodelistVersion GetCodelistVersion(EA.Package selectedPgk)
        {
            CodelistVersion codelistVersion = null;

            foreach(EA.Element elem in selectedPgk.Elements)
            {
                if(elem.Type.ToLower().Equals("class") && elem.Name.ToLower().Equals("version"))
                {
                    codelistVersion = new CodelistVersion();
                    foreach (EA.Attribute attr in elem.Attributes)
                    {
                        if (attr.Name.ToLower().Equals("version")) codelistVersion.Version = attr.Default;
                        if (attr.Name.ToLower().Equals("title")) codelistVersion.Title = attr.Default;
                        if (attr.Name.ToLower().Equals("release")) codelistVersion.Release = attr.Default;
                        if (attr.Name.ToLower().Equals("releasedate")) codelistVersion.ReleaseDate = attr.Default;
                    }
                }
            }
            return codelistVersion;
        }


        public string GetOutputFileName()
        {
            return xmlParser.GetXmlValueJsonData("LastJsonExportedFile");
        }

        public string GetCodelistOutputFileName()
        {
            string dir = xmlParser.GetXmlValueJsonData(DialCodelistExportToJson.LASTJSONCODELIST_EXPORT_DIRECTORY);
            string filename = xmlParser.GetXmlValueJsonData(DialCodelistExportToJson.LASTJSONCODELIST_EXPORT_FILENAME);
            if(!dir.EndsWith("\\")) dir += "\\";
            return dir + filename;
        }

        public string GetCodelistTemplateOutputFileName(DialogResponse dialogResponseExpTmplJson)
        {
            string dir = dialogResponseExpTmplJson.DirName;
            string filename = dialogResponseExpTmplJson.FileName;
            if (!dir.EndsWith("\\")) dir += "\\";
            return dir + filename;
        }


        private List<JsonBase> GetRootElements(List<EA.Element> rootElements)
        {
            List<JsonBase> resList = new List<JsonBase>();
            JsonObject<JsonBase> jPropsObj = new JsonObject<JsonBase>("properties");
            List<string> requiredElements = new List<string>();
            foreach (EA.Element rootElem in rootElements)
            {
                JsonObject<JsonBase> jRootObj = new JsonObject<JsonBase>(rootElem.Name);
                List<JsonKeyValue<string>> minMaxItems = JsonGenerationUtilities.GetRootMinMax(rootElem);
                if (minMaxItems != null && minMaxItems.Count > 0) // root element is array
                {
                    JsonKeyValue<string> jType = new JsonKeyValue<string>(SCHEMA_TYPE, "array");
                    JsonObject<JsonBase> jItems = new JsonObject<JsonBase>("items");
                    JsonKeyValue<string> jRootRef = new JsonKeyValue<string>(SCHEMA_ELEM_REF, "#/" + SCHEMA_ELEM_DEFS + "/" + rootElem.Name);
                    jItems.Add(jRootRef);
                    jRootObj.Add(jType);
                    jRootObj.Add(jItems);
                    foreach (JsonKeyValue<string> jKeyValue in minMaxItems)
                    {
                        jRootObj.Add(jKeyValue);
                    }
                }
                else
                {
                    requiredElements.Add(rootElem.Name);
                    JsonKeyValue<string> jRootRef = new JsonKeyValue<string>(SCHEMA_ELEM_REF, "#/" + SCHEMA_ELEM_DEFS + "/" + rootElem.Name);
                    jRootObj.Add(jRootRef);
                }
                jPropsObj.Add(jRootObj);
            }
            resList.Add(jPropsObj);

            // required elements
            if (requiredElements.Count > 0)
            {
                requiredElements.Sort();
                JsonArray<string> jRequired = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                foreach (string requElemName in requiredElements)
                {
                    jRequired.Add(requElemName);
                }
                resList.Add(jRequired);
            }
            return resList;
        }

        private List<JsonBase> GetAllSubSchemas(List<EA.Element> allElems)
        {
            List<JsonBase> subSchemas = new List<JsonBase>();
            try
            {
                foreach (EA.Element elem in allElems)
                {
                    if (elem.Name == null || elem.Name.Length == 0) continue;

                    JsonBase jBase = null;

                    if (elem.Type.Equals("Enumeration") || (elem.Stereotype.ToLower().Equals("enumeration")))
                    {
                        jBase = MapEnumeration(elem);
                    }
                    else if (elem.Type.Equals("Class"))
                    {
                        if (elem.Stereotype == null || elem.Stereotype.Equals(""))
                        {
                            if (elem.Abstract.Equals("1")) return null;
                            jBase = MapClass(elem);
                        }
                        else if (elem.StereotypeEx.Contains("enumeration") || elem.Stereotype.Contains("Enumeration"))
                        {
                            jBase = MapEnumeration(elem);
                        }
                        else if (elem.Stereotype.Contains("Compound"))
                        {
                            jBase = MapCompound(elem);
                        }
                        else if (elem.Stereotype.Contains("CIMDatatype"))
                        {
                            jBase = MapDatatype(elem);
                        }
                        else if (elem.Stereotype.Contains("Primitive"))
                        {
                            jBase = MapPrimitive(elem);
                        }
                        else
                        {
                            jBase = MapClass(elem);
                        }
                    }
                    if (jBase == null)
                    {
                        logger.Log(elem.Name);
                        logger.Log("Stereotype: '" + elem.Stereotype + "'");
                    } else
                    {
                        subSchemas.Add(jBase);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message + ": \n" + ex.StackTrace);
            }
            
            JsonComparer jsonComparer = new JsonComparer();
            subSchemas.Sort(jsonComparer);
            return subSchemas;
        }

        public List<JsonBase> CreateBasicParams(EA.Element elem)
        {
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_ELEM_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>("title", elem.Name);

            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);

            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_ADD_PROPS, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));

            List<JsonBase> jKList = new List<JsonBase>();
            jKList.Add(jKVId);
            jKList.Add(jKVTitle);
            jKList.Add(jKVDescription);
            jKList.Add(jKVModelReference);
            return jKList;
        }

        public List<JsonBase> CreateGeneralParams(EA.Element elem)
        {
            List<JsonBase> jKList = CreateBasicParams(elem);
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(SCHEMA_TYPE, "object");
            JsonKeyValue<bool>   jKVAddProperties = new JsonKeyValue<bool>(SCHEMA_ADD_PROPS, false);
            jKList.Add(jKVType);
            jKList.Add(jKVAddProperties);
            return jKList;
        }


        public JsonObject<JsonBase> MapEnumeration(EA.Element elem)
        {
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(SCHEMA_TYPE, "string");
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_DEF_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(SCHEMA_DEF_TITLE, elem.Name);

            string listElems = "";
            if (elem.StereotypeEx.Contains("CodeList"))
            {
                foreach (EA.Attribute attr in elem.Attributes)
                {
                    listElems = (listElems.Length <= 0) ? (listElems + attr.Default) : (listElems + ", " + attr.Default);
                }
            }
            else
            {
                foreach (EA.Attribute attr in elem.Attributes)
                {
                    listElems = (listElems.Length <= 0) ? (listElems + attr.Name) : (listElems + ", " + attr.Name);
                }
            }
            JsonKeyValue<string> jKVEnums = new JsonKeyValue<string>("enum", "[" + listElems + "]");

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(elem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            jObj.Add(jKVType);
            jObj.Add(jKVEnums);
            return jObj;
        }

        public JsonObject<JsonBase> MapCompound(EA.Element elem)
        {
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(SCHEMA_TYPE, "string");
            JsonKeyValue<bool> jKVAddProperties = new JsonKeyValue<bool>(SCHEMA_ADD_PROPS, false);
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_DEF_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(SCHEMA_DEF_TITLE, elem.Name);

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(elem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            jObj.Add(jKVType);
            jObj.Add(jKVAddProperties);

            List<string> requiredAttributes = new List<string>();
            // properties
            JsonObject<JsonBase> jProps = new JsonObject<JsonBase>("properties");
            foreach (EA.Attribute attr in elem.Attributes)
            {
                if (attr.LowerBound != null && attr.LowerBound.Trim().Length > 0 && !attr.LowerBound.Equals("0"))
                {
                    requiredAttributes.Add(attr.Name);
                }
                JsonObject<JsonBase> mappedProp;
                if (JsonGenerationUtilities.IsBasicType(attr.Name))
                {
                    mappedProp = this.jpMapper.MapSimpleProperty(elem.Name, attr);
                } else
                {
                    mappedProp = this.jpMapper.MapCompoundProperty(elem.Name, attr);
                }
                jProps.Add(mappedProp);
            }
            jObj.Add(jProps);

            JsonArray<string> jArray = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
            foreach (string requAttr in requiredAttributes)
            {
                jArray.Add(requAttr);
            }
            jObj.Add(jArray);

            return jObj;
        }

        public JsonObject<JsonBase> MapDatatype(EA.Element elem)
        {
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            string name = elem.Name;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + name));
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(SCHEMA_TYPE, "object");
            JsonKeyValue<bool> jKVAddProperties = new JsonKeyValue<bool>(SCHEMA_ADD_PROPS, false);
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_DEF_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(SCHEMA_DEF_TITLE, elem.Name);

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(elem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            jObj.Add(jKVType);
            jObj.Add(jKVAddProperties);

            JsonObject<JsonBase> jProps = new JsonObject<JsonBase>("properties");
            List<JsonObject<JsonBase>> mappedProperties = new List<JsonObject<JsonBase>>(); 
            List<string> requiredAttributes = new List<string>();
            foreach (EA.Attribute attr in elem.Attributes)
            {
                JsonObject<JsonBase> mappedProp = this.jpMapper.MapDatatypeProperty(elem.Name, attr);
                if (jpMapper.GetCodelistPack() != null) this.codelistPackage = jpMapper.GetCodelistPack();
                
                if (attr.Name.Equals("value"))
                {
                    mappedProperties.Insert(0, mappedProp); // put value on first place
                } else
                {
                    mappedProperties.Add(mappedProp);
                }
                if (attr.LowerBound != null && attr.LowerBound.Trim().Length > 0 && !attr.LowerBound.Equals("0"))
                {
                    requiredAttributes.Add(attr.Name);
                }
            }

            foreach (JsonObject<JsonBase> mappedProp in mappedProperties)
            {
                jProps.Add(mappedProp);
            }
            jObj.Add(jProps); 

            JsonArray<string> jArray = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
            foreach(string requAttr in requiredAttributes)
            {
                jArray.Add(requAttr);
            }
            jObj.Add(jArray);
            return jObj;
        }

        public JsonObject<JsonBase> MapPrimitive(EA.Element elem)
        {
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_DEF_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(SCHEMA_DEF_TITLE, elem.Name);
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);

            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));

            List<JsonKeyValue<string>> basicTypeParams = JsonGenerationUtilities.GetBasicType(elem.Name);
            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(elem.Type);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            foreach (JsonKeyValue<string> jkv in basicTypeParams)
            {
                jObj.Add(jkv);
            }
            return jObj;
        }


        public JsonObject<JsonBase> MapClass(EA.Element elem)
        {
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(SCHEMA_TYPE, "object");
            JsonKeyValue<bool> jKVAddProperties = new JsonKeyValue<bool>(SCHEMA_ADD_PROPS, false);
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(SCHEMA_DEF_ID, "#" + elem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(SCHEMA_DEF_TITLE, elem.Name);

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(elem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            jObj.Add(jKVType);
            jObj.Add(jKVAddProperties);

            JsonObject<JsonBase> jProps = new JsonObject<JsonBase>(SCHEMA_ELEM_PROPERTIES);
            List<string> requiredAttributes = new List<string>();
            // proper class
            foreach (EA.Attribute attr in elem.Attributes)
            {
                if (attr.LowerBound != null && attr.LowerBound.Trim().Length > 0 && !attr.LowerBound.Equals("0"))
                {
                    requiredAttributes.Add(attr.Name);
                }
                JsonObject<JsonBase> mappedProp = this.jpMapper.MapSimpleProperty(elem.Name, attr);
                jProps.Add(mappedProp);
            }

            /*-------------------------------------------------------------------------

            // inheritance chain
            List<EA.Element> inheritanceChain = new List<EA.Element>();
            JsonGenerationUtilities.GetInheritedChain(this.repo, elem, inheritanceChain);
            foreach (EA.Element chainElem in inheritanceChain)
            {
                foreach (EA.Attribute attr in chainElem.Attributes)
                {
                    JsonObject<JsonBase> mappedProp = MapSimpleProperty(chainElem.Name, attr);
                    jProps.Add(mappedProp);
                }
            }
            -------------------------------------------------------------------------------*/
            // connectors
            JsonArray<JsonBase> jAllOf = null; // for required properties
            foreach (EA.Connector conn in elem.Connectors)
            {
                if (conn.Type.Equals("Dependency") || conn.Type.Equals("Generalization"))
                {
                    continue;
                }

                EA.ConnectorEnd theOtherConnEnd;
                EA.Element otherEndElem;
                long aggregation; // Indicates aggregation or composition or association
                                    // of always the other end!
                if (elem.ElementID == conn.ClientID)
                {
                    aggregation = conn.SupplierEnd.Aggregation;
                    theOtherConnEnd = conn.SupplierEnd;
                    otherEndElem = repo.GetElementByID(conn.SupplierID);
                }
                else
                {
                    aggregation = conn.ClientEnd.Aggregation;
                    theOtherConnEnd = conn.ClientEnd;
                    otherEndElem = repo.GetElementByID(conn.ClientID);
                }

                if (aggregation > 0) // 1 .. Aggregation, 2 .. Composition
                {
                    continue;
                }

                List<EA.Element> unionSubClasses = GetUnionSubclasses(repo, otherEndElem);
                if (unionSubClasses.Count > 1)
                {
                    Dictionary<string, string> unionSubClassMapping = GetUnionClassMapping(theOtherConnEnd, otherEndElem, unionSubClasses);
                    foreach (EA.Element subElem in unionSubClasses)
                    {
                        JsonObject<JsonBase> mappedProp = this.jpMapper.MapTypedProperty(subElem, theOtherConnEnd, requiredAttributes, unionSubClassMapping);
                        jProps.Add(mappedProp);
                    }
                    jAllOf = GetRequiredPropertiesOfUnion(theOtherConnEnd, unionSubClassMapping, unionSubClasses);
                }
                else
                {
                    JsonObject<JsonBase> mappedProp = this.jpMapper.MapTypedProperty(otherEndElem, theOtherConnEnd, requiredAttributes, null);
                    jProps.Add(mappedProp);
                }

                //if (!theOtherConnEnd.Navigable.Equals("Navigable")) continue; TO CHECK
            }

            //*---- XOR ---------------------------------------------------------------------
            // If element is not an xor owner:
            List<XOROwner> xorOwnerships = GetXOROwnerships(xorOwnerList, elem);
            if (xorOwnerships.Count <= 0) // no XOR ownership for this element
            {
                jProps.Sort();
                jObj.Add(jProps);
                JsonArray<string> jRequ = JsonGenerationUtilities.GetRequiredProperties(requiredAttributes);
                if(jRequ != null)
                {
                    jObj.Add(jRequ);
                }
            } else
            {
                jProps.Sort();
                jObj.Add(jProps);
                // If element is an xor owner:
                JsonArray<JsonBase> jAllOfXOR = GetAllOfXOR(xorOwnerships, requiredAttributes);
                if(jAllOf != null)
                {
                    jAllOf.Append(jAllOfXOR.GetList());
                } else
                {
                    jAllOf = jAllOfXOR;
                }
            }

            if(jAllOf != null)
            {
                jObj.Add(jAllOf);
            }
            return jObj;
        }

        private JsonArray<JsonBase> GetRequiredPropertiesOfUnion(EA.ConnectorEnd theOtherConnEnd, Dictionary<string, string> unionSubClassMapping, List<EA.Element> unionSubClasses)
        {
            JsonArray<JsonBase> jAllOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ALLOF);
            int minCard = JsonGenerationUtilities.GetMinCardinality(theOtherConnEnd.Cardinality);
            int maxCard = JsonGenerationUtilities.GetMaxCardinality(theOtherConnEnd.Cardinality);
            if((minCard == maxCard) && (minCard == 1)) // union-props-constrain-min-one-max-one
            {
                JsonObject<JsonBase> jOneOfHull = GetUnionPropsConstrainMinOneMaxOne(unionSubClassMapping, unionSubClasses);
                jAllOf.Add(jOneOfHull);
                return jAllOf;
            } else if((minCard == 0) && (maxCard == 1)) // union-props-constrain-min-zero-max-one
            {
                JsonObject<JsonBase> jOneOfOverallHull = GetUnionPropsConstrainMinZeroMaxOne(unionSubClassMapping, unionSubClasses);
                jAllOf.Add(jOneOfOverallHull);
                return jAllOf;
            }
            else if((minCard == 1) && (maxCard > 1)) // union-props-constrain-min-one-max-greater-than-one
            {
                JsonArray<JsonBase> jAnyOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ANYOF);
                foreach (EA.Element subClass in unionSubClasses)
                {
                    string mappedName = unionSubClassMapping[subClass.Name];
                    JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                    jRequ.Add(mappedName);
                    JsonObject<JsonBase> jHull = new JsonObject<JsonBase>("");
                    jHull.Add(jRequ);
                    jAnyOf.Add(jHull);
                }
                JsonObject<JsonBase> jOneOfHull = GetUnionPropsConstrainMinOneMaxGreaterThanOne(unionSubClassMapping, unionSubClasses);
                jAllOf.Add(jOneOfHull);
                return jAllOf;
            }
            return null;
        }

        private JsonObject<JsonBase> GetUnionPropsConstrainMinOneMaxGreaterThanOne(Dictionary<string, string> unionSubClassMapping, List<EA.Element> unionSubClasses)
        {
            JsonArray<JsonBase> jAnyOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ANYOF);
            foreach (EA.Element subClass in unionSubClasses)
            {
                string mappedName = unionSubClassMapping[subClass.Name];
                JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                jRequ.Add(mappedName);
                JsonObject<JsonBase> jHull = new JsonObject<JsonBase>("");
                jHull.Add(jRequ);
                jAnyOf.Add(jHull);
            }
            JsonObject<JsonBase> jOneOfHull = new JsonObject<JsonBase>("");
            jOneOfHull.Add(jAnyOf);
            return jOneOfHull;
        }
        private JsonObject<JsonBase> GetUnionPropsConstrainMinZeroMaxOne(Dictionary<string, string> unionSubClassMapping, List<EA.Element> unionSubClasses)
        {
            JsonArray<JsonBase> jOneOfOverall = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
            JsonArray<JsonBase> jOneOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
            JsonArray<JsonBase> jAllOfNot = new JsonArray<JsonBase>(SCHEMA_ELEM_ALLOF);
            foreach (EA.Element subClass in unionSubClasses)
            {
                string mappedName = unionSubClassMapping[subClass.Name];
                JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                jRequ.Add(mappedName);
                JsonObject<JsonBase> jHull = new JsonObject<JsonBase>("");
                jHull.Add(jRequ);
                jOneOf.Add(jHull);
                JsonObject<JsonBase> jOneOfHull = new JsonObject<JsonBase>("");
                jOneOfHull.Add(jOneOf); // OneOf-Required part

                JsonArray<JsonBase> jNot = new JsonArray<JsonBase>(SCHEMA_ELEM_NOT);
                jNot.Add(jHull);
                JsonObject<JsonBase> jNotHull = new JsonObject<JsonBase>("");
                jNotHull.Add(jNot); // Not-Required part
                jAllOfNot.Add(jNotHull);
                JsonObject<JsonBase> jAllOfNotHull = new JsonObject<JsonBase>("");
                jAllOfNotHull.Add(jAllOfNot); // AllOf-Not-Required part

                jOneOfOverall.Add(jOneOfHull);
                jOneOfOverall.Add(jAllOfNotHull);
            }
            JsonObject<JsonBase> jOneOfOverallHull = new JsonObject<JsonBase>("");
            jOneOfOverallHull.Add(jOneOfOverall);
            return jOneOfOverallHull;
        }
        private JsonObject<JsonBase> GetUnionPropsConstrainMinOneMaxOne(Dictionary<string, string> unionSubClassMapping, List<EA.Element> unionSubClasses)
        {
            JsonArray<JsonBase> jOneOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
            JsonObject<JsonBase> jOneOfHull = new JsonObject<JsonBase>("");
            foreach (EA.Element subClass in unionSubClasses)
            {
                string mappedName = unionSubClassMapping[subClass.Name];
                JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                jRequ.Add(mappedName);
                JsonObject<JsonBase> jHull = new JsonObject<JsonBase>("");
                jHull.Add(jRequ);
                jOneOf.Add(jHull);
            }
            jOneOfHull.Add(jOneOf);
            return jOneOfHull;
        }

        private Dictionary<string, string> GetUnionClassMapping(EA.ConnectorEnd theOtherConnEnd, EA.Element otherEndElement, List<EA.Element> unionSubClasses)
        {
            Dictionary<string, string> unionSubClassMapping = new Dictionary<string, string>();
            string qualifier;
            if (theOtherConnEnd.Role.Equals(otherEndElement.Name)) { // role identical to class name
                foreach(EA.Element elem in unionSubClasses)
                {
                    unionSubClassMapping.Add(elem.Name, elem.Name);
                }
            } else if((qualifier = GetQualifier(theOtherConnEnd.Role, otherEndElement.Name)) != null) // role name has qualifier
            {
                foreach (EA.Element elem in unionSubClasses)
                {
                    unionSubClassMapping.Add(elem.Name, qualifier + "_" + elem.Name);
                }
            }
            else
            {
                foreach (EA.Element elem in unionSubClasses)
                {
                    unionSubClassMapping.Add(elem.Name, theOtherConnEnd.Role + "_" + elem.Name);
                }
            }

            return unionSubClassMapping;
        }

        private string GetQualifier(string roleName, string unionClassName)
        {
            int index = roleName.IndexOf('_');
            if((index >= 0) && (roleName.Substring(index + 1).Equals(unionClassName)))
            {
                return roleName.Substring(0, index);
            }
            return null;
        }

        private List<EA.Element> GetUnionSubclasses(EA.Repository repo, EA.Element otherEndElement)
        {
            List<EA.Element> subClassesList = new List<EA.Element>();
            foreach(EA.Connector conn in otherEndElement.Connectors)
            {
                if(conn.Type.Equals("Generalization") && 
                  (otherEndElement.StereotypeEx != null) && 
                  (otherEndElement.StereotypeEx.ToLower().Contains(STEREOTYPE_UNION)))
                {
                    if(conn.SupplierID == otherEndElement.ElementID) // is superclass
                    {
                        subClassesList.Add(repo.GetElementByID(conn.ClientID));
                    }
                }
            }
            return subClassesList;
        }

        private JsonArray<JsonBase> GetAllOfXOR(List<XOROwner> xorOwnerships, List<string> requiredAttributes)
        {
            JsonArray<JsonBase> jAllOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ALLOF);
            foreach (XOROwner xorOwner in xorOwnerships)
            {
                List<string>[] requOptLists = JsonGenerationUtilities.GetRequiredLists(repo, xorOwner);
                if (requOptLists[OPTIONAL].Count <= 0) // just required ones
                {
                    JsonArray<JsonBase> jOneOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
                    foreach (string otherElemName in requOptLists[REQUIRED])
                    {
                        JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                        jRequ.Add(otherElemName);
                        JsonObject<JsonBase> jRequHull = new JsonObject<JsonBase>("");
                        jRequHull.Add(jRequ);
                        jOneOf.Add(jRequHull);
                    }
                    JsonObject<JsonBase> jOneOfHull = new JsonObject<JsonBase>("");
                    jOneOfHull.Add(jOneOf);
                    JsonArray<string> jRequProps = JsonGenerationUtilities.GetRequiredProperties(requiredAttributes);
                    if (jRequProps != null)
                    {
                        JsonObject<JsonBase> jHull = new JsonObject<JsonBase>("");
                        jHull.Add(jRequProps);
                        jAllOf.Add(jHull);
                    }
                    jAllOf.Add(jOneOfHull);
                }
                else // mixed
                {
                    List<string> allTogether = new List<string>();
                    allTogether.AddRange(requOptLists[REQUIRED]);
                    allTogether.AddRange(requOptLists[OPTIONAL]);
                    JsonArray<JsonBase> jOneOfMixed = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
                    foreach (string otherElemName in allTogether)
                    {
                        JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                        jRequ.Add(otherElemName);
                        JsonObject<JsonBase> jRequHull = new JsonObject<JsonBase>("");
                        jRequHull.Add(jRequ);
                        jOneOfMixed.Add(jRequHull);
                    }
                    JsonObject<JsonBase> jOneOfMixedHull = new JsonObject<JsonBase>("");
                    jOneOfMixedHull.Add(jOneOfMixed);

                    JsonArray<JsonBase> jAllOfMixed = new JsonArray<JsonBase>(SCHEMA_ELEM_ALLOF);
                    foreach (string otherElemName in allTogether)
                    {
                        JsonArray<string> jRequ = new JsonArray<string>(SCHEMA_ELEM_REQUIRED);
                        jRequ.Add(otherElemName);
                        JsonObject<JsonBase> jRequHull = new JsonObject<JsonBase>("");
                        jRequHull.Add(jRequ);
                        JsonArray<JsonBase> jNot = new JsonArray<JsonBase>(SCHEMA_ELEM_NOT);
                        jNot.Add(jRequHull);
                        JsonObject<JsonBase> jNotHull = new JsonObject<JsonBase>("");
                        jNotHull.Add(jNot);
                        jAllOfMixed.Add(jNotHull);
                    }
                    JsonObject<JsonBase> jAllOfMixedHull = new JsonObject<JsonBase>("");
                    jAllOfMixedHull.Add(jAllOfMixed);

                    JsonArray<JsonBase> jOneOf = new JsonArray<JsonBase>(SCHEMA_ELEM_ONEOF);
                    jOneOf.Add(jOneOfMixedHull);
                    jOneOf.Add(jAllOfMixedHull);
                    JsonObject<JsonBase> jOneOfHull = new JsonObject<JsonBase>("");
                    jOneOfHull.Add(jOneOf);
                    jAllOf.Add(jOneOfHull);
                }
            }
            return jAllOf;
        }

        private List<XOROwner> GetXOROwnerships(List<JsonMaster.XOROwner> xorOwnerList, EA.Element elem)
        {
            List<XOROwner> ownershipList = new List<XOROwner>();
            foreach (XOROwner xorOwner in xorOwnerList)
            {
                if (xorOwner.GetOwner().ElementID == elem.ElementID)
                {
                    ownershipList.Add(xorOwner);
                }
            }
            return ownershipList;
        }

    }
}
