﻿using CimSyntaxGen.Option;
using CimSyntaxGen.Utilities;
using EA;
using System;
using System.Collections.Generic;
using System.Diagnostics.Eventing.Reader;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CimSyntaxGen.JsonGeneration
{
    public class JsonPropertiesMapper
    {
        private Logger logger = new Logger(Main.LOG_FILE, Main.DO_LOG);
        private string rootModelURI;
        private EA.Repository repo;
        private string schemaType;
        private XMLParser xmlParser;
        private EA.Package codelistPack = null;

        public JsonPropertiesMapper(string rootModelURI, EA.Repository repo, string schemaType, XMLParser xmlParser) 
        {
            this.rootModelURI = rootModelURI;
            this.repo = repo;
            this.schemaType = schemaType;
            this.xmlParser = xmlParser;
        }

        public JsonPropertiesMapper(EA.Repository repo) 
        {
            this.repo = repo;
        }

        private string GetRelativePath(string localCodelistFileName, string outputDir)
        {
            if(localCodelistFileName.Contains(outputDir))
            {
                return localCodelistFileName.Replace(outputDir, ""); // create relative path
            }
            return localCodelistFileName;
        }

        public JsonObject<JsonBase> MapCodelist(EA.Element codelistElem, string localCodelistFileName, string modelRefUrl, string outputDir)
        {
            string notes = (codelistElem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(codelistElem.Notes) : "";
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_ID, "#" + codelistElem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_TITLE, codelistElem.Name);
            if (!modelRefUrl.EndsWith("#")) { modelRefUrl += "#"; }
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(JsonMapper.SCHEMA_MODEL_REF, modelRefUrl + codelistElem.Name);
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, notes);

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(codelistElem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);

            JsonArray<JsonBase> oneOf = new JsonArray<JsonBase>(JsonMapper.SCHEMA_ELEM_ONEOF);
            JsonObject<JsonBase> jStd = new JsonObject<JsonBase>(null);
            JsonKeyValue<string> jKVStd = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/Standard" + codelistElem.Name);
            jStd.Add(jKVStd);
            oneOf.Add(jStd);
            JsonObject<JsonBase> jLoc = new JsonObject<JsonBase>(null);
            string localFileName = GetRelativePath(localCodelistFileName, outputDir);
            JsonKeyValue<string> jKVLoc = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF,
                localFileName.Replace("\\", "/") + "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/Local" + codelistElem.Name);
            jLoc.Add(jKVLoc);
            oneOf.Add(jLoc);

            jObj.Add(oneOf);
            return jObj;
        }


        public JsonObject<JsonBase> MapStandardCodelist(EA.Element codelistElem)
        {
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_ID, "#Standard" + codelistElem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_TITLE, "");
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(JsonMapper.SCHEMA_TYPE, "string");

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>("Standard" + codelistElem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVType);

            JsonArray<string> jEnums = new JsonArray<string>("enum");

            /* ABA20240112 -> duplicate 'description' keys in JSON are a problem!
            foreach(EA.Attribute attr in codelistElem.Attributes)
            {
                jEnums.Add(attr.Default);
                string description = attr.Default + " ... '" + attr.Name + "': " + attr.Notes.Trim().Replace("\"", "'").Replace("\n", "").Replace("\r", "");
                JsonKeyValue<string> desc = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, description);
                jObj.Add(desc);
            }
            */

            StringBuilder description = new StringBuilder();
            foreach (EA.Attribute attr in codelistElem.Attributes)
            {
                jEnums.Add(attr.Default);
                if (description.Length > 0) { description.Append("\\n"); }
                description.Append(attr.Default + " ... '" + attr.Name + "': " + attr.Notes.Trim().Replace("\"", "'").Replace("\n", "").Replace("\r", ""));
            }
            JsonKeyValue<string> desc = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, description.ToString());
            jObj.Add(desc);

            jObj.Add(jEnums);
            return jObj;
        }

        public JsonObject<JsonBase> MapLocalCodelist(EA.Element codelistElem)
        {
            JsonKeyValue<string> jKVId = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_ID, "#Local" + codelistElem.Name);
            JsonKeyValue<string> jKVTitle = new JsonKeyValue<string>(JsonMapper.SCHEMA_DEF_TITLE, "");
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, "");
            JsonKeyValue<string> jKVType = new JsonKeyValue<string>(JsonMapper.SCHEMA_TYPE, "string");

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>("Local" + codelistElem.Name);
            jObj.Add(jKVId);
            jObj.Add(jKVTitle);
            jObj.Add(jKVDescription);
            jObj.Add(jKVType);

            JsonArray<string> jEnums = new JsonArray<string>("enum");
            jEnums.Add("A01");
            jObj.Add(jEnums);
            return jObj;
        }


        public JsonObject<JsonBase> MapSimpleProperty(string parentElemName, EA.Attribute attr)
        {
            string notes = (attr.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(attr.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            string name = parentElemName + "." + attr.Name;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(JsonMapper.SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + name));
            List<JsonKeyValue<string>> basicTypeParams = JsonGenerationUtilities.GetBasicType(attr.Type);

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(attr.Name);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);
            if (basicTypeParams != null && basicTypeParams.Count > 0)
            {
                foreach (JsonKeyValue<string> kv in basicTypeParams)
                {
                    jObj.Add(kv);
                }
            }
            else
            {
                JsonKeyValue<string> jTypeRef = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + attr.Type);
                jObj.Add(jTypeRef);
            }
            return jObj;
        }

        public JsonObject<JsonBase> MapTypedProperty(EA.Element elem, EA.ConnectorEnd otherConnEnd, List<string> requiredAttributes, Dictionary<string, string> unionSubClassMapping)
        {
            string notes = (elem.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(elem.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(JsonMapper.SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + elem.Name));
            JsonObject<JsonBase> jObj;
            if (unionSubClassMapping != null)
            {
                jObj = new JsonObject<JsonBase>(unionSubClassMapping[elem.Name]);
            }
            else
            {
                jObj = new JsonObject<JsonBase>(elem.Name);
            }
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);

            int minCard = JsonGenerationUtilities.GetMinCardinality(otherConnEnd.Cardinality);
            int maxCard = JsonGenerationUtilities.GetMaxCardinality(otherConnEnd.Cardinality);
            if (maxCard > 1 || maxCard == -1) // "-1"  stand for arbitraty many
            {
                JsonKeyValue<string> jKVType = new JsonKeyValue<string>(JsonMapper.SCHEMA_TYPE, "array");
                jObj.Add(jKVType);
                JsonObject<JsonBase> jItems = new JsonObject<JsonBase>("items");
                JsonKeyValue<string> jTypeRef = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + elem.Name);
                jItems.Add(jTypeRef);
                jObj.Add(jItems);
                JsonKeyValue<int> jMin = new JsonKeyValue<int>(JsonMapper.MIN_ITEMS, minCard);
                jObj.Add(jMin);
                if(maxCard > 0) 
                {
                    JsonKeyValue<int> jMax = new JsonKeyValue<int>(JsonMapper.MAX_ITEMS, maxCard);
                    jObj.Add(jMax);
                }
            }
            else 
            {
                JsonKeyValue<string> jKVType = new JsonKeyValue<string>(JsonMapper.SCHEMA_TYPE, "object");
                jObj.Add(jKVType);
                JsonKeyValue<string> jTypeRef = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + elem.Name);
                jObj.Add(jTypeRef);
            }
            if (minCard >= 1)
            {
                requiredAttributes.Add(elem.Name);
            }
            return jObj;
        }

        private string GetCodelistRelativeFilePath()
        {
            string jsonCodeListFilename = xmlParser.GetXmlValueJsonData(DialCodelistExportToJson.LASTJSONCODELIST_EXPORT_FILENAME);
            string jsonCodeListDir = xmlParser.GetXmlValueJsonData(DialCodelistExportToJson.LASTJSONCODELIST_EXPORT_DIRECTORY);
            if (!jsonCodeListDir.EndsWith("\\")) { jsonCodeListDir += "\\"; }

            string lastJsonExportedFullFilename = xmlParser.GetXmlValueJsonData(Dialjson.LAST_JSON_EXPORTED_FILE);
            string outputDir = Path.GetDirectoryName(lastJsonExportedFullFilename);
            if (!outputDir.EndsWith("\\")) { outputDir += "\\"; }

            string relFilePath = jsonCodeListFilename;
            if(jsonCodeListDir.Contains(outputDir))
            {
                relFilePath = jsonCodeListDir.Replace(outputDir, "") + jsonCodeListFilename;
            }
            return relFilePath.Replace("\\", "/");
        }

        public JsonObject<JsonBase> MapDatatypeProperty(string parentElemName, EA.Attribute attr)
        {
            string notes = (attr.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(attr.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, notes);
            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(attr.Name);
            jObj.Add(jKVDescription);

            List<JsonKeyValue<string>> jBasicType = JsonGenerationUtilities.GetBasicType(attr.Type);
            if (jBasicType.Count <= 0) // no basic type
            {
                if (IsEnumeration(attr))
                {
                    string defaultValue = null;
                    try
                    {
                        defaultValue = attr.Default;
                    }
                    catch (Exception) { /* ignore */ }
                    JsonKeyValue<string> jTypeRef = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + attr.Type);
                    if (defaultValue == null || defaultValue.Equals(""))
                    {
                        jObj.Add(jTypeRef);
                    }
                    else
                    {
                        JsonKeyValue<string> jConst;
                        if (attr.IsConst)
                        {
                            jConst = new JsonKeyValue<string>("const", defaultValue);
                        }
                        else
                        {
                            jConst = new JsonKeyValue<string>("default", defaultValue);
                        }
                        JsonObject<JsonBase> jAllOfObj = new JsonObject<JsonBase>(JsonMapper.SCHEMA_ELEM_ALLOF);
                        jAllOfObj.Add(jTypeRef);
                        jAllOfObj.Add(jConst);
                        jObj.Add(jAllOfObj);
                    }
                }
                else if (IsTypeESMPCodeList(attr))
                {
                    string eUrn = GetCodelistRelativeFilePath();
                    string refString = eUrn + "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + attr.Type;
                    JsonKeyValue<string> jKVReference = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, refString);
                    jObj.Add(jKVReference);

                    string defaultValue = null;
                    try
                    {
                        defaultValue = attr.Default;
                    }
                    catch (Exception) { /* ignore */ }
                    if (defaultValue != null && defaultValue.Trim().Length > 0)
                    {
                        JsonKeyValue<string> jConst;
                        if (attr.IsConst)
                        {
                            jConst = new JsonKeyValue<string>("const", defaultValue);
                        }
                        else
                        {
                            jConst = new JsonKeyValue<string>("default", defaultValue);
                        }
                        jObj.Add(jConst);
                    }
                }
            }
            else
            {
                if (jBasicType.Count == 1)
                {
                    jObj.Add(jBasicType[0]);
                }
                else
                {
                    jObj.Add(jBasicType[0]);
                    jObj.Add(jBasicType[1]);
                }
            }

            List<JsonKeyValue<string>> jKVList = GetMinMax(attr);
            foreach (JsonKeyValue<string> jKVElement in jKVList)
            {
                jObj.Add(jKVElement);
            }
            return jObj;
        }

        private bool IsTypeESMPCodeList(EA.Attribute attr)
        {
            EA.Element classifier = this.repo.GetElementByID(attr.ClassifierID);
            string classifierName = classifier.Name.ToLower();
            if (classifierName.Contains("typelist") ||
                classifierName.Contains("unitsymbol") ||
                classifierName.Contains("unitmultiplier")) 
            {
                this.codelistPack = repo.GetPackageByID(classifier.PackageID);
                if (this.codelistPack != null) logger.Log("coelistPack = " + codelistPack.Name);
                else logger.Log("package of classifier " + classifierName + " is null!");
                return true; 
            }
            return false;
        }

        public EA.Package GetCodelistPack() { return this.codelistPack; }

        private bool IsEnumeration(EA.Attribute attr)
        {
            EA.Element classifier = this.repo.GetElementByID(attr.ClassifierID);
            string classifierName = classifier.Name.ToLower();
            if ((schemaType.Equals(JsonMapper.SCHEMA_TYPE_WG16)) && 
                (classifierName.Contains("typelist") || 
                 classifierName.Contains("unitsymbol") || 
                 classifierName.Contains("unitmultiplier"))) return false;
            else if (classifier.Type.ToLower().Equals("enumeration") || 
                     classifier.StereotypeEx.ToLower().Contains("enumeration")) return true;
            return false;
        }

        public List<JsonKeyValue<string>> GetMinMax(EA.Attribute attr)
        {
            List<JsonKeyValue<string>> jKVList = new List<JsonKeyValue<string>>();
            string min = "";
            string max = "";
            if (attr.LowerBound != null && attr.LowerBound.Length > 0) min = attr.LowerBound;
            if (attr.UpperBound != null && attr.UpperBound.Length > 0) max = attr.UpperBound;
            if (!min.Equals(max))
            {
                if (min.Length > 0)
                {
                    JsonKeyValue<string> jKVMin = new JsonKeyValue<string>(JsonMapper.SCHEMA_MIN, attr.LowerBound);
                    jKVList.Add(jKVMin);
                }
                if (max.Length > 0)
                {
                    JsonKeyValue<string> jKVMax = new JsonKeyValue<string>(JsonMapper.SCHEMA_MAX, attr.UpperBound);
                    jKVList.Add(jKVMax);
                }
            }
            return jKVList;
        }

        public JsonObject<JsonBase> MapCompoundProperty(string parentElemName, EA.Attribute attr)
        {
            string notes = (attr.Notes != null) ? JsonGenerationUtilities.ReplaceLFCR(attr.Notes) : "";
            JsonKeyValue<string> jKVDescription = new JsonKeyValue<string>(JsonMapper.SCHEMA_DESC, notes);
            string modelRef = this.rootModelURI;
            string name = parentElemName + "." + attr.Name;
            JsonKeyValue<string> jKVModelReference = new JsonKeyValue<string>(JsonMapper.SCHEMA_MODEL_REF, JsonGenerationUtilities.RemoveQual(modelRef + name));

            JsonObject<JsonBase> jObj = new JsonObject<JsonBase>(attr.Name);
            jObj.Add(jKVDescription);
            jObj.Add(jKVModelReference);

            JsonKeyValue<string> jTypeRef = new JsonKeyValue<string>(JsonMapper.SCHEMA_ELEM_REF, "#/" + JsonMapper.SCHEMA_ELEM_DEFS + "/" + attr.Name);
            jObj.Add(jTypeRef);
            return jObj;
        }


    }
}
